"""
Cluster upstream sequences (UTR and leader) for each gene

For each gene, look at the sequences assigned to it. Take the upstream
sequences and compute a consensus for them. Only those assigned sequences are
taken that have a very low error rate for the V gene match.

Output a FASTA file that contains one consensus sequence for each gene.
"""
import logging
from collections import Counter

from ..table import read_table
from ..align import iterative_consensus

logger = logging.getLogger(__name__)

# When computing a UTR consensus, ignore sequences that deviate by more than
# this factor from the median length.
UTR_MEDIAN_DEVIATION = 0.1


def add_arguments(parser):
    arg = parser.add_argument
    arg('--max-V-errors', '--max-error-percentage', '-e', dest='max_v_errors',
        metavar='PERCENT', type=float, default=1,
        help='Allow PERCENT errors in V gene match. Default: %(default)s')
    arg('--max-FR1-errors', dest='max_fr1_errors', metavar='PERCENT', type=float, default=None,
        help='Allow PERCENT errors in FR1 region.')
    arg('--max-CDR1-errors', dest='max_cdr1_errors', metavar='PERCENT', type=float, default=None,
        help='Allow PERCENT errors in CDR1 region.')
    arg('--min-consensus-size', type=int, default=1, metavar='N',
        help='Require at least N sequences for consensus. Default: %(default)s')
    arg('--consensus-threshold', '-t', metavar='PERCENT',
        type=float, default=75,
        help='Threshold for consensus computation. Default: %(default)s%%')
    arg('--no-ambiguous', '--no-N', default=False, action='store_true',
        help='Discard consensus sequences with ambiguous bases')
    arg('--part', choices=['UTR', 'leader', 'UTR+leader'],
        default='UTR+leader', help='Which part of the sequence before the V '
        'gene match to analyze. Default: %(default)s')
    arg('--debug', default=False, action='store_true',
        help='Enable debugging output')
    arg('table', help='Table with parsed IgBLAST results (assigned.tab.gz or '
        'filtered.tab.gz)')


def main(args):
    if args.debug:
        logging.getLogger().setLevel(logging.DEBUG)
    table = read_table(args.table)
    n_genes = len(set(table['V_gene']))
    logger.info('%s rows read with %d unique gene names', len(table), n_genes)
    for name, column_name, arg in (
            ('V%SHM', 'V_SHM', args.max_v_errors),
            ('FR1%SHM', 'FR1_SHM', args.max_fr1_errors),
            ('CDR1%SHM', 'CDR1_SHM', args.max_cdr1_errors)):
        if arg is not None:
            table = table[getattr(table, column_name) <= arg]
            logger.info('%s rows remain after discarding %s > %s%%',
                len(table), name, arg)
    table['UTR+leader'] = table['UTR'] + table['leader']
    table = table[table[args.part] != '']
    table['UTR_length'] = [ len(s) for s in table['UTR'] ]

    n_written = 0
    n_consensus_with_n = 0
    for name, group in table.groupby('V_gene'):
        assert len(group) != 0
        if len(group) < args.min_consensus_size:
            logger.info('Gene %s has too few assignments (%s), skipping.', name, len(group))
            continue

        counter = Counter(group['UTR_length'])
        logger.debug('Sequence length/count table: %s',
            ', '.join('{}: {}'.format(l, c) for l, c in counter.most_common()))

        if args.part == 'leader':
            sequences = list(group['leader'])
        else:
            # Since UTR lengths vary a lot, take all sequences that are at
            # least as long as the tenth longest sequence, and compute
            # consensus only from those.
            length_threshold = sorted(group['UTR_length'], reverse=True)[:10][-1]
            sequences = list(group[group['UTR_length'] >= length_threshold][args.part])

        if len(sequences) == 0:
            logger.info('Gene %s has %s assignments, but lengths are too different, skipping.', name, len(group))
            continue
        assert len(sequences) > 0
        if len(sequences) == 1:
            cons = sequences[0]
        else:
            # Keep only those sequences whose length is at least 90% of the longest one
            cons = iterative_consensus(sequences, program='muscle-medium', threshold=args.consensus_threshold/100)
            # If the sequence lengths are different, the beginning can be unclear
            cons = cons.lstrip('N')
        logger.info('Gene %s has %s assignments, %s usable (%s unique sequences). '
            'Consensus has %s N bases.', name, len(group), len(sequences),
            len(set(sequences)), cons.count('N'))
        if cons.count('N') > 0:
            n_consensus_with_n += 1
            if args.no_ambiguous:
                continue
        n_written += 1
        print('>{} {}_consensus\n{}'.format(name, args.part, cons))

    in_or_ex = 'excluding' if args.no_ambiguous else 'including'
    logger.info('Wrote a consensus for %s of %s genes (%s %s with ambiguous bases)', n_written, n_genes, in_or_ex, n_consensus_with_n)



"""
Run IgDiscover

Calls Snakemake to produce all the output files.
"""
import sys
import logging
import resource
import platform
import pkg_resources
from snakemake import snakemake

from ..utils import available_cpu_count
from .. import __version__

from .config import Config

logger = logging.getLogger(__name__)


def add_arguments(parser):
    arg = parser.add_argument
    arg('--dryrun', '-n', default=False, action='store_true',
        help='Do not execute anything')
    arg('--cores', '--jobs', '-j', metavar='N', type=int, default=available_cpu_count(),
        help='Run on at most N CPU cores in parallel. '
        'Default: Use as many cores as available (%(default)s)')
    arg('--keepgoing', '-k', default=False, action='store_true',
        help='If one job fails, finish the others.')
    arg('targets', nargs='*', default=[],
        help='File(s) to create. If omitted, the full pipeline is run.')


def main(args):
    try:
        config = Config.from_default_path()
    except FileNotFoundError as e:
        sys.exit("Pipeline configuration file {!r} not found. Please create it!".format(e.filename))

    print('IgDiscover version {} with Python {}. Configuration:'.format(__version__,
        platform.python_version()))
    for k, v in sorted(vars(config).items()):
        # TODO the following line is only necessary for non-YAML configurations
        if k.startswith('_'):
            continue
        print('   ', k, ': ', repr(v), sep='')
    sys.stdout.flush()

    # snakemake sets up its own logging and this cannot be easily changed
    # (setting keep_logger=True crashes), so remove our own log handler
    # for now
    logger.root.handlers = []
    snakefile_path = pkg_resources.resource_filename('igdiscover', 'Snakefile')
    success = snakemake(
        snakefile_path,
        snakemakepath='snakemake',  # Needed in snakemake 3.9.0
        dryrun=args.dryrun,
        cores=args.cores,
        keepgoing=args.keepgoing,
        printshellcmds=True,
        targets=args.targets if args.targets else None,
    )

    if sys.platform == 'linux' and not args.dryrun:
        cputime = resource.getrusage(resource.RUSAGE_SELF).ru_utime
        cputime += resource.getrusage(resource.RUSAGE_CHILDREN).ru_utime
        h = int(cputime // 3600)
        m = (cputime - h * 3600) / 60
        print('Total CPU time: {}h {:.2f}m'.format(h, m))

    sys.exit(0 if success else 1)



"""
This trie can be used to store a set of strings and to retrieve sequences similar to a query
sequence. 
"""


class Trie:
    """
    A tree-like datastructure for storing strings. It supports queries similar to a set() of
    strings. This particular implementation also supports searching for strings by Hamming distance.
    """
    # Children
    A = None
    C = None
    G = None
    T = None
    # The name attribute is set to the string that is spelled from the root of the tree to this
    # node - but only for leaf nodes.
    name = None

    def __init__(self, iterable=None):
        if iterable is not None:
            for it in iterable:
                self.add(it)

    def add(self, s: str):
        """Add a string to this trie. If it already exists, the trie remains unchanged."""
        self._insert(s, leaf_name=s)

    def _insert(self, s: str, leaf_name: str):
        """Recursive insert, called by add()"""
        if len(s) == 0:
            # This needs to become a leaf node
            assert self.name is None or self.name == leaf_name
            self.name = leaf_name
        else:
            subtrie = getattr(self, s[0])
            if subtrie is None:
                subtrie = Trie()
                setattr(self, s[0], subtrie)
            subtrie._insert(s[1:], leaf_name)

    def __repr__(self):
        parts = []
        for c in list('ACGT'):
            if getattr(self, c) is not None:
                if getattr(self, c) is None:
                    parts.append(c)
                else:
                    parts.append(c + ':' + repr(getattr(self, c)))
        if self.name is not None:
            parts.append('Leaf({})'.format(self.name))
        if parts:
            return '{' + ', '.join(parts) + '}'
        else:
            return '*'

    def find_node(self, s: str):
        """
        Search for s. If s is a prefix of any string in the trie, then return the node
        corresponding to that prefix. The node can be an internal or a leaf node.
        A leaf node is a node in which name attribute is not None. If the returned node
        is a leaf node, then s is in the trie. If it is an internal node (that is,
        the name attribute is None), then there exists a string in the trie that has
        s as a proper prefix.
        
        If no such string (with s as prefix) exists, then None is returned.
        """
        if len(s) == 0:
            return self
        subtrie = getattr(self, s[0])
        if subtrie is None:
            return None
        else:
            return subtrie.find_node(s[1:])

    def __contains__(self, s: str):
        """Return whether s is in the trie"""
        node = self.find_node(s)
        return node is not None and node.name == s

    def count_nodes(self, internal=True):
        """Return number of nodes in this trie. internal nodes are counted if "internal" is True"""
        n = 1 if internal or self.name is not None else 0
        for c in list('ACGT'):
            subtrie = getattr(self, c)
            if subtrie is not None:
                n += len(subtrie)
        return n

    def __len__(self):
        """Return the number of unique strings in this trie"""
        return self.count_nodes(internal=False)

    def has_similar(self, s, mismatches):
        """
        Return whether a string exists in this trie that has a Hamming distance
        of at most 'mismatches' to s.
        """
        if len(s) == 0:
            return self.name is not None
        # As a runtime heuristic, descend into the the mismatches==0 subtrie first
        # if possible.
        subtrie = getattr(self, s[0])
        if subtrie is not None and subtrie.has_similar(s[1:], mismatches):
            return True
        if mismatches == 0:
            return False
        # The above did not work - try all three possible substitutions
        # and recursively check subtries
        for c in list('ACGT'):
            if c == s[0]:
                continue
            subtrie = getattr(self, c)
            if subtrie is None:
                continue
            if subtrie.has_similar(s[1:], mismatches - 1):
                return True
        return False

    def find_all_similar(self, s, mismatches):
        """
        Yield all strings in this trie that have a Hamming distance of at most
        'mismatches' to s.
        """
        # This routine is similar to has_similar, but since we are interested
        # in all similar strings, the optimization of going into the matching
        # subtrie first does not help.
        if len(s) == 0:
            if self.name is not None:
                yield self.name
            return

        # The code below is an optimized version of the following:
        #
        # for c in list('ACGT'):
        #   if c != s[0] and mismatches == 0:
        #     continue
        #   subtrie = getattr(self, c)
        #   if subtrie is not None:
        #     yield from subtrie.find_all_similar(s[1:], mismatches - int(c != s[0]))

        if mismatches > 0:
            s0 = s[0]
            s1 = s[1:]
            subtrie = self.A
            if subtrie is not None:
                yield from subtrie.find_all_similar(s1, mismatches - (int(s0 != 'A')))
            subtrie = self.C
            if subtrie is not None:
                yield from subtrie.find_all_similar(s1, mismatches - (int(s0 != 'C')))
            subtrie = self.G
            if subtrie is not None:
                yield from subtrie.find_all_similar(s1, mismatches - (int(s0 != 'G')))
            subtrie = self.T
            if subtrie is not None:
                yield from subtrie.find_all_similar(s1, mismatches - (int(s0 != 'T')))
        else:
            subtrie = self
            for c in s:
                subtrie = getattr(subtrie, c)
                if subtrie is None:
                    break
            else:
                if subtrie.name is not None:
                    yield subtrie.name



import os
import sys
import pytest

from igdiscover.__main__ import main
from .utils import datapath, resultpath, files_equal


@pytest.fixture
def run(tmpdir):
    def _run(args, expected):
        """
        Run IgDiscover, redirecting stdout to a temporary file.
        Then compare the output with the contents of an expected file.
        """
        outpath = str(tmpdir.join('output'))
        print('Running:', ' '.join(args))
        with open(outpath, 'w') as f:
            old_stdout = sys.stdout
            sys.stdout = f
            main(args)
            sys.stdout = old_stdout
        assert files_equal(expected, outpath)

    return _run


def test_main():
    with pytest.raises(SystemExit) as exc:
        main(['--version'])
    assert exc.value.code == 0


def test_group_by_barcode_only(run):
    args = ['group', '-b', '4', datapath('ungrouped.fasta')]
    run(args,  resultpath('grouped-by-barcode-only.fasta'))


def test_group_by_pseudo_cdr3(run):
    args = ['group', '-b', '4', '--pseudo-cdr3=-5:-2', '--trim-g', datapath('ungrouped.fasta')]
    run(args,  resultpath('grouped.fasta'))


def test_group_by_pseudo_cdr3_barcode_at_end(run):
    args = ['group', '-b', '-4', '--pseudo-cdr3=1:3', datapath('ungrouped.fasta')]
    run(args, resultpath('grouped2.fasta'))


def test_clusterplot(tmpdir):
    main(['clusterplot', '-m', '10', datapath('clusterplot.tab.gz'), str(tmpdir)])
    assert tmpdir.join('IGHV1-1801.png').check()


def test_igblast(run):
    args = ['igblast', '--threads=1', datapath('database/'), datapath('igblast.fasta')]
    run(args, resultpath('assigned.tab'))



from collections import Counter
import dnaio

import matplotlib.pyplot as plt
import numpy as np


def length_histogram(path):
    """Return a list of lengths """
    lengths = []
    with dnaio.open(path) as reader:
        for record in reader:
            lengths.append(len(record.sequence))
    return lengths


def plot_histogram(lengths, path, title, min_x=0, bins=50):
    """
    Plot histogram of lengths to path
    """
    lengths = np.array(lengths)
    histomax = int(max(lengths, default=100))
    fig = plt.figure(figsize=(20/2.54, 10/2.54))
    ax = fig.gca()
    ax.set_xlabel('Read length')
    ax.set_ylabel('Frequency')
    ax.set_title(title)
    _, borders, _ = ax.hist(lengths, bins=bins, range=(min_x, histomax))
    fig.set_tight_layout(True)
    fig.savefig(path)


def read_length_histogram(path, tsv_path, plot_path, bins, left, title):
    lengths = length_histogram(path)
    freqs = Counter(lengths)
    with open(tsv_path, "w") as outfile:
        print("## File:", path, file=outfile)
        print("length", "frequency", sep='\t', file=outfile)
        for length in range(0, max(freqs, default=100) + 1):
            freq = freqs[length]
            if freq > 0:
                print(length, freq, sep='\t', file=outfile)

    if plot_path:
        plot_histogram(lengths, plot_path, title, min_x=left, bins=bins)



from igdiscover.cluster import inner_nodes, hamming_single_linkage


def inner_nodes_recursive(root):
    """
    Return a list of all inner nodes of the tree, from left to right.
    """
    if root.is_leaf():
        return []
    return inner_nodes_recursive(root.left) + [root] + inner_nodes_recursive(root.right)


def collect_ids_recursive(root):
    """
    Return a list of ids of all leaves of the given tree
    """
    if root.is_leaf():
        return [root.id]
    return collect_ids_recursive(root.left) + collect_ids_recursive(root.right)


def test_inner_nodes():
    class Node:
        def __init__(self, value, left, right):
            self.value = value
            self.left = left
            self.right = right

        def is_leaf(self):
            return self.left is None and self.right is None

        def __repr__(self):
            # return 'Node({!r}, {!r}, {!r})'.format(self.value, self.left, self.right)
            return 'Node({!r}, ...)'.format(self.value)

    empty_tree = None
    assert inner_nodes(empty_tree) == []

    leaf = Node(0, None, None)
    assert inner_nodes(leaf) == []

    tree = Node(1, leaf, leaf)
    values = [ n.value for n in inner_nodes(tree) ]
    assert values == [1]

    tree = Node(1, Node(2, Node(3, leaf, leaf), leaf), leaf)
    values = [ n.value for n in inner_nodes(tree) ]
    assert values == [3, 2, 1]

    tree = Node(1, Node(2, Node(3, leaf, Node(4, leaf, leaf)), leaf), leaf)
    values = [ n.value for n in inner_nodes(tree) ]
    assert values == [3, 4, 2, 1]

    tree = Node(1,
        leaf,
        Node(2,
            leaf,
            Node(3,
                Node(4,
                    leaf,
                    Node(5, leaf, leaf)
                ),
                Node(6, leaf, leaf)
            )
        )
    )
    values = [ n.value for n in inner_nodes(tree) ]
    assert values == [1, 2, 4, 5, 3, 6]


def test_hamming_single_linkage():
    strings = ['', 'AC', 'AG', 'GG', 'CT', 'GGG', 'GGA']
    components = hamming_single_linkage(strings, 0)
    assert all(len(component) == 1 for component in components)
    assert set(strings) == set(component[0] for component in components)

    components = hamming_single_linkage(strings, 1)
    components = set(frozenset(component) for component in components)
    expected = [[''], ['AC', 'AG', 'GG'], ['CT'], ['GGG', 'GGA']]
    expected = set(frozenset(component) for component in expected)
    assert components == expected



"""
Compare two FASTA files based on sequences

The order of records in the two files does not matter.

Exit code:
    2 if duplicate sequences or duplicate record names were found
    1 if there are any lost or gained records or sequence differences
    0 if the records are identical, but allowing for different record names

"""
import sys
import logging
import numpy as np
from scipy.optimize import linear_sum_assignment
import dnaio
from tinyalign import hamming_distance

logger = logging.getLogger(__name__)

do_not_show_cpustats = 1


def add_arguments(parser):
    arg = parser.add_argument
    arg('--color', default='auto', choices=('auto', 'never', 'always'),
        help='Whether to colorize output')
    arg('a', help='FASTA file with expected sequences')
    arg('b', help='FASTA file with actual sequences')


RED = "\x1b[0;31m"
GREEN = "\x1b[0;32m"
RESET = "\x1b[0m"


def red(s):
    return RED + s + RESET


def green(s):
    return GREEN + s + RESET


def check_duplicate_names(records):
    names = set()
    for record in records:
        if record.name in names:
            yield record.name
        names.add(record.name)


def check_exact_duplicate_sequences(records):
    sequences = dict()
    for record in records:
        if record.sequence in sequences:
            yield record.name, sequences[record.sequence]
        else:
            sequences[record.sequence] = record.name


def compare(a, b):
    """Return cost of comparing a to b"""

    l = min(len(a.sequence), len(b.sequence))
    length_diff = max(len(a.sequence), len(b.sequence)) - l
    dist_prefixes = hamming_distance(a.sequence[:l], b.sequence[:l])
    dist_suffixes = hamming_distance(a.sequence[-l:], b.sequence[-l:])

    return 5 * min(dist_prefixes, dist_suffixes) + length_diff


def pair_up_identical(a_records, b_records):
    identical = []
    b_map = {record.sequence: record for record in b_records}
    a_rest = []
    for a in a_records:
        if a.sequence in b_map:
            identical.append((a, b_map[a.sequence]))
            del b_map[a.sequence]
        else:
            a_rest.append(a)

    return identical, a_rest, list(b_map.values())


def pair_up(a_records, b_records, max_cost=20):
    # Pair up identical sequences first
    identical, a_records, b_records = pair_up_identical(a_records[:], b_records[:])

    # Compare all vs all and fill in a score matrix
    m = len(a_records)
    n = len(b_records)
    cost = np.zeros((m, n), dtype=int)
    for i, a in enumerate(a_records):
        for j, b in enumerate(b_records):
            cost[i, j] = compare(a, b)

    # Solve minimum weighted bipartite matching
    assignment = linear_sum_assignment(cost)
    similar = []
    a_similar = set()
    b_similar = set()
    for i, j in zip(*assignment):
        if cost[i, j] <= max_cost:
            similar.append((a_records[i], b_records[j]))
            a_similar.add(i)
            b_similar.add(j)

    a_only = [a for i, a in enumerate(a_records) if i not in a_similar]
    b_only = [b for j, b in enumerate(b_records) if j not in b_similar]

    return a_only, b_only, identical, similar


def format_indel(a, b, colored: bool):
    if len(a) > len(b):
        assert len(b) == 0
        s = '{-' + a + '}'
        return red(s) if colored else s
    elif len(b) > len(a):
        assert len(a) == 0
        s = '{+' + b + '}'
        return green(s) if colored else s
    else:
        return ''


def print_similar(a, b, colored: bool):
    l = min(len(a.sequence), len(b.sequence))
    dist_prefixes = hamming_distance(a.sequence[:l], b.sequence[:l])
    dist_suffixes = hamming_distance(a.sequence[-l:], b.sequence[-l:])
    if dist_prefixes <= dist_suffixes:
        a_prefix = ''
        b_prefix = ''
        a_common = a.sequence[:l]
        b_common = b.sequence[:l]
        a_suffix = a.sequence[l:]
        b_suffix = b.sequence[l:]
    else:
        a_prefix = a.sequence[:-l]
        b_prefix = b.sequence[:-l]
        a_common = a.sequence[-l:]
        b_common = b.sequence[-l:]
        a_suffix = ''
        b_suffix = ''

    s = format_indel(a_prefix, b_prefix, colored)
    edits = []
    for i, (ac, bc) in enumerate(zip(a_common, b_common)):
        if ac != bc:
            if colored:
                s = '{' + red(ac) + ' → ' + green(bc) + '}'
            else:
                s = '{' + ac + ' → ' + bc + '}'
            edits.append(s)
        else:
            edits.append(ac)
    s += ''.join(edits)

    s += format_indel(a_suffix, b_suffix, colored)

    print('~', a.name, '--', b.name)
    print(s)
    print()


def main(args):
    if args.color == 'auto':
        colored = sys.stdout.isatty()
    elif args.color == 'never':
        colored = False
    else:
        assert args.color == 'always'
        colored = True

    with dnaio.open(args.a) as f:
        a_records = list(f)
    with dnaio.open(args.b) as f:
        b_records = list(f)

    has_duplicate_names = False
    for records, path in ((a_records, args.a), (b_records, args.b)):
        dups = list(check_duplicate_names(records))
        if dups:
            has_duplicate_names = True
            print('Duplicate record names found in', path)
            for name in dups:
                print('-', name)

    has_duplicate_sequences = False
    for record, path in ((a_records, args.a), (b_records, args.b)):
        dups = list(check_exact_duplicate_sequences(records))
        if dups:
            has_duplicate_sequences = True
            print('Duplicate sequences found in', path)
            for name, name_orig in dups:
                print('-', name, 'is identical to earlier record', name_orig)

    only_a, only_b, identical, similar = pair_up(a_records, b_records)
    different_name = [(a, b) for a, b in identical if a.name != b.name]

    # Summary
    print('{} vs {} records. {} lost, {} gained, {} identical, {} different name, {} similar'.format(
        len(a_records), len(b_records), len(only_a), len(only_b),
        len(identical) - len(different_name), len(different_name),
        len(similar)))

    # Report what has changed
    if only_a:
        print()
        print('## Only in A')
        for record in only_a:
            print('-', record.name)
    if only_b:
        print()
        print('## Only in B')
        for record in only_b:
            print('+', record.name)
    if different_name:
        print()
        print('## Different name (sequence identical)')
        for a, b in different_name:
            print('=', a.name, '--', b.name)
    if similar:
        print()
        print('## Similar')
        for a, b in similar:
            print_similar(a, b, colored)

    if has_duplicate_names or has_duplicate_sequences:
        sys.exit(2)
    if only_a or only_b or similar:
        sys.exit(1)
    # different name is fine for success
    sys.exit(0)



# Tests for CDR3 detection
from igdiscover.species import find_cdr3
from igdiscover.utils import nt_to_aa


def split(s):
    for line in s.split('\n'):
        line = line.strip()
        if line:
            yield line.split()


def assert_cdr3_detection(chain, s):
    for amino_acids, sequence in split(s):
        for offset in range(3):
            target = sequence[offset:]
            match = find_cdr3(target, chain)
            assert match is not None
            assert nt_to_aa(target[match[0]:match[1]]) == amino_acids, (chain, amino_acids, offset)


def test_cdr3_detection_heavy():
    heavy = """
    ARRLHSGSYILFDY CAGGTGACCTTGAAGGAGTCTGGTCCTGCGCTGGTGAAACCCACACAGACCCTCACGCTGACCTGCACCTTCTCTGGGTTCTCACTCAGCACTAGTGGTATGGGTGTGGGCTGGATCCGTCAGCCCTCACGGAAGACCCTGGAGTGGCTTGCACACATTTATTGGAATGATGATAAATACTACAGCACATCGCTGAAGAGCAGGCTCACCATCTCCAAGGACACCTCCAAAAACCAGGTGGTTCTAACAATGACCAACATGGACCCTGTGGACACAGCCACATATTACTGTGCACGGAGACTTCATAGTGGGAGCTACATTCTCTTTGACTACTGGGGCCAGGGAGTCCTGGTCACCGTCTCCTCAGGGAGTGCATCCGCCCCAACCCTTTTCCCCCTCGTCTCCTGTGA

    ARIKWLRSPGYGYFDF CAGGTGACCTTGAAGGAGTCTGGTCCTGCGCTGGTGAGACCCACACAGACCCTCACTCTGACCTGCACCTTCTCTGGGTTCTCAATCAGCACCTCTGGAACAGGTGTGGGCTGGATCCGTCAGCCCCCAGGGAAGGCCCTGGAATGGCTTGCAAGCATTTATTGGACTGATGCTAAATACTATAGCACATCGCAGAAGAGCAGGCTCACCATCTCCAAGGACACCTCCAGAAACCAGGTGATTCTAACAATGACCAACATGGAGCCTGTGGACATAGCCACATATTTCTGTGCACGGATAAAGTGGCTGCGGTCCCCAGGCTATGGATACTTCGATTTCTGGGGCCCTGGCACCCCAATCACCATCTCCTCAGGGAGTGCATCCGCCCCAACCCTTTTCCCCCTCGTCTCCTGTGA

    ARHGIAAAGTHNWFDP TCAGCCGACAAGTCCATCAGCACCGCCTACCTGCAGTGGAGCAGCCTGAAGGCCTCGGACACCGCCATGGATTACTGTGCGAGACATGGGATAGCAGCAGCTGGTACCCACAACTGGTTCGACCCCTGGGGCCAGGGAACCCTGGTCACCGTCTCCTCAGGGAGTGCATCCGCCCCAACCCTTTTCCCCCTCGTCTCCTGTGAGAATTCCCCGTCGGCAGGTTGTT
    """
    assert_cdr3_detection('VH', heavy)


def test_cdr3_detection_kappa():
    kappa = """
QQYDSSPRT TTCAGTGGCAGTGGAGCAGGGACAGATTTCACTCTCACCATCAGCAGTCTGGAACCTGAGGATGTCGCAACTTACTACTGTCAGCAGTATGATAGCAGCCCCCGGACGTTCGGCGCTGGGACCAAGCTGGAAATCAAACGGAGTGTGCAGAAGCCAACTATCTCCCTCTTCCCTCCATCATCTGAGGAGG

QQYSSYPYT GAGCTGGCCTCGGGAGTCCCAGCTCGCTTCAGTGGGAGTGGGTCAGGGACTTCTTTCTCTCTCACAATCAGCAACGTGGAGGCTGAAGATGTTGCAACCTATTACTGTCAGCAGTATAGCAGTTATCCGTACACGTTCGGCGCAGGGACCAAGCTGGAAATCAAACGGAGTGTGCAGAAGCCAACTATCTCCCTCTTCCCTCCATCATCTGAGGAGG

LQYDSSPYT ATTCTCACCATCAGCAGCCTGCAGCCTGAAGACTTTGCAACTTACTACTGTCTACAGTATGATAGTTCCCCGTACACGTTCGGCGCAGGGACCAAGCTGGAAATCAAACGGAGTGTGCAGAAGCCAACTATCTCCCTCTTCCCTCCATCATCTGAGGAGG

FQYYSGRLT ACAGACTTCACTCTCACCATCAGCAGCCTGCAGCCTGAGGACATTGCAGTTTATTACTGTTTCCAGTATTACAGCGGGAGACTCACGTTCGGAGGAGGGACCCGCTTGGAAATCAAACGGAGTGTGCAGAAGCCAACTATCTCCCTCTTCCCTCCATCATCTGAGGAGG
    """
    assert_cdr3_detection('VK', kappa)


def test_cdr3_detection_lambda():
    lambda_ = """
LTYHGNSGTFV GGATCCAAAAACCCCTCAGCCAATGCAGGAATTTTGCTCATCTCTGAACTCCAGAATGAGGATGAGGCTGACTATTACTGTCTGACATATCATGGTAATAGTGGTACTTTTGTATTCGGTGGAGGAACCAAGCTGACCGTCCTAGGTCAGCCCAAGTCTGCCCCCACAGTCAGCCTGTTCC

QLWDANSTV ATGGCCACACTGACCATCACTGGCGCCCAGGGTGAGGACGAGGCCGACTATTGCTGTCAGTTGTGGGATGCTAACAGTACTGTGTTCGGTGGAGGAACCACGCTGACCGTCCTAGGTCAGCCCAAGTCTGCCCCCACAGTCAGCCTGTTCCCGCCCTCCTC

GVGYSGGYV GATCGCTACTTAACCATCTCCAACATCCAGCCTGAAGACGAGGCTGACTATTTCTGTGGTGTGGGTTATAGCGGTGGTTATGTATTCGGTGGAGGAACCAAGTTGACCGTCCTAGGTCAGCCCAAGTCTGCTCCCACAGTCAGCCTGTTCCCGCCCTCCTC
    """
    assert_cdr3_detection('VL', lambda_)



"""
Change configuration file

Example:
    igdiscover config --set iterations 1

If no --set option is given, print the current configuration.
"""
import os
import logging
import ruamel.yaml

from igdiscover.config import Config

logger = logging.getLogger(__name__)


def add_arguments(parser):
    arg = parser.add_argument
    arg('--set', nargs=2, default=[], metavar=('KEY', 'VALUE'), action='append',
        help='Set KEY to VALUE. Use KEY.SUBKEY[.SUBSUBKEY...] for nested keys.'
            'Can be given multiple times.')
    arg('--file', default=Config.DEFAULT_PATH,
        help='Configuration file to modify. Default: igdiscover.yaml in current directory.')


def main(args):
    if args.set:
        with open(args.file) as f:
            config = ruamel.yaml.load(f, ruamel.yaml.RoundTripLoader)
        for k, v in args.set:
            v = ruamel.yaml.safe_load(v)
            # config[k] = v
            item = config
            # allow nested keys
            keys = k.split('.')
            for i in keys[:-1]:
                item = item[i]
            item[keys[-1]] = v
        tmpfile = args.file + '.tmp'
        with open(tmpfile, 'w') as f:
            print(ruamel.yaml.dump(config, Dumper=ruamel.yaml.RoundTripDumper), end='', file=f)
        os.rename(tmpfile, args.file)
    else:
        with open(args.file) as f:
            config = ruamel.yaml.safe_load(f)
        print(ruamel.yaml.dump(config), end='')



#!/usr/bin/env python3
from cutadapt import seqio
import sys
from collections import defaultdict

with seqio.open(sys.argv[1]) as sr:
	reads = list(sr)

d = defaultdict(list)
for read in reads:
	d[read.sequence].append(read.name)

for seq, names in d.items():
	print('>{}\n{}'.format(','.join(names), seq))




"""
Plot histograms of differences to reference V gene

For each gene, a histogram is plotted that shows how often a sequence was
assigned to that gene at a certain percentage difference.
"""
import sys
import logging
import numpy as np
import seaborn as sns
from matplotlib.backends.backend_pdf import FigureCanvasPdf, PdfPages
from matplotlib.figure import Figure

from ..table import read_table

logger = logging.getLogger(__name__)


def add_arguments(parser):
    arg = parser.add_argument
    arg('--minimum-group-size', '-m', metavar='N', default=None, type=int,
        help="Plot only genes with at least N assigned sequences. "
        "Default: 0.1%% of assigned sequences or 100, whichever is smaller.")
    arg('--max-j-shm', metavar='VALUE', type=float, default=None,
        help='Use only rows with J%%SHM >= VALUE')
    arg('--multi', metavar='PDF', default=None,
        help='Plot individual error frequency histograms (for each V gene) to this PDF file')
    arg('--boxplot', metavar='PDF', default=None,
        help='Plot a single page with box(en)plots of V SHM for multiple genes')
    arg('table', metavar='FILTERED.TAB.GZ', help='Table with parsed IgBLAST results')


def plot_difference_histogram(group, gene_name, bins=np.arange(20.1)):
    """
    Plot a histogram of percentage differences for a specific gene.
    """
    exact_matches = group[group.V_SHM == 0]
    cdr3s_exact = len(set(s for s in exact_matches.CDR3_nt if s))
    js_exact = len(set(exact_matches.J_gene))

    fig = Figure(figsize=(100/25.4, 60/25.4))
    ax = fig.gca()
    ax.set_xlabel('Percentage difference')
    ax.set_ylabel('Frequency')
    fig.suptitle('Gene ' + gene_name, y=1.08, fontsize=16)
    ax.set_title('{:,} sequences assigned'.format(len(group)))

    ax.text(0.25, 0.95,
        '{:,} ({:.1%}) exact matches\n  {} unique CDR3\n  {} unique J'.format(
            len(exact_matches), len(exact_matches) / len(group),
            cdr3s_exact, js_exact),
        transform=ax.transAxes, fontsize=10,
        bbox=dict(boxstyle='round', facecolor='white', alpha=0.5),
        horizontalalignment='left', verticalalignment='top')
    _ = ax.hist(list(group.V_SHM), bins=bins)
    return fig


def main(args):
    table = read_table(args.table, usecols=['V_gene', 'J_gene', 'V_SHM', 'J_SHM', 'CDR3_nt'])
    if not args.multi and not args.boxplot:
        print('Don’t know what to do', file=sys.stderr)
        sys.exit(2)

    # Discard rows with any mutation within J at all
    logger.info('%s rows read', len(table))
    if args.max_j_shm is not None:
        # Discard rows with too many J mutations
        table = table[table.J_SHM <= args.max_j_shm][:]
        logger.info('%s rows remain after keeping only those with J%%SHM <= %s',
            len(table), args.max_j_shm)

    if args.minimum_group_size is None:
        total = len(table)
        minimum_group_size = min(total // 1000, 100)
        logger.info('Skipping genes with less than %s assignments', minimum_group_size)
    else:
        minimum_group_size = args.minimum_group_size

    # Genes with high enough assignment count
    all_genes = table['V_gene'].unique()
    genes = sorted(table['V_gene'].value_counts().loc[lambda x: x >= minimum_group_size].index)
    gene_set = set(genes)

    logger.info('%s out of %s genes have enough assignments', len(genes), len(all_genes))
    if args.multi:
        with PdfPages(args.multi) as pages:
            for gene, group in table.groupby('V_gene'):
                if gene not in gene_set:
                    continue
                fig = plot_difference_histogram(group, gene)
                FigureCanvasPdf(fig).print_figure(pages, bbox_inches='tight')

        logger.info('Wrote %r', args.multi)

    if args.boxplot:
        aspect = 1 + len(genes) / 32
        g = sns.catplot(x='V_gene', y='V_SHM', kind='boxen', order=genes, data=table,
            height=2 * 2.54, aspect=aspect, color='g')
        # g.set(ylim=(-.1, None))
        g.set(ylabel='% V SHM (nt)')
        g.set(xlabel='V gene')
        g.set_xticklabels(rotation=90)
        g.savefig(args.boxplot)
        logger.info('Wrote %r', args.boxplot)



"""
Some utility functions that work on sequences and lists of sequences.
"""
import os
import sys
import re
import random
import hashlib
import resource
from collections import defaultdict
from typing import List

import dnaio
import numpy as np
from tinyalign import edit_distance
from cutadapt.align import Aligner

from .dna import GENETIC_CODE


def downsampled(population, size):
    """
    Return a random subsample of the population.

    Uses reservoir sampling. See https://en.wikipedia.org/wiki/Reservoir_sampling
    """
    sample = population[:size]
    for index in range(size, len(population)):
        r = random.randint(0, index)
        if r < size:
            sample[r] = population[index]
    return sample


def distances(sequences, band=0.2):
    """
    Compute all pairwise edit distances and return a square matrix.

    Entry [i,j] in the matrix is the edit distance between sequences[i]
    and sequences[j].
    """
    # Pre-compute distances between unique sequences
    unique_sequences = list(set(sequences))
    unique_distances = dict()  # maps (seq1, seq2) tuples to edit distance
    maxdiff = max((int(len(s) * band) for s in sequences), default=0)  # TODO double-check this
    for i, s in enumerate(unique_sequences):
        for j, t in enumerate(unique_sequences):
            if i < j:
                dist = min(maxdiff+1, edit_distance(s, t, maxdiff=maxdiff))
                unique_distances[(t, s)] = dist
                unique_distances[(s, t)] = dist

    # Fill the result matrix
    m = np.zeros((len(sequences), len(sequences)), dtype=float)
    for i, s in enumerate(sequences):
        for j, t in enumerate(sequences):
            if i < j:
                d = 0 if s == t else unique_distances[(s, t)]
                m[j, i] = m[i, j] = d
    return m


def sequence_hash(s, digits=4):
    """
    For a string, return a 'fingerprint' that looks like 'S1234' (the character
    'S' is fixed). The idea is that this allows one to quickly see whether two
    sequences are not identical.
    """
    h = int(hashlib.md5(s.encode()).hexdigest()[-4:], base=16)
    return 'S' + str(h % 10**digits).rjust(digits, '0')


def unique_name(name, sequence):
    """Create a unique name based on the current name and the sequence

    The returned name looks like name_S1234. If the current name contains
    already a S_.... suffix, it is removed before the new suffix is appended.

    name -- current name
    """
    return '{}_{}'.format(name.rsplit('_S', 1)[0], sequence_hash(sequence))


class SerialPool:
    """
    An alternative to multiprocessing.Pool that runs things in serial for
    easier debugging
    """
    def __init__(self, *args, **kwargs):
        pass

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass

    def imap(self, func, iterable, chunksize):
        for i in iterable:
            yield func(i)


def natural_sort_key(s, _nsre=re.compile('([0-9]+)')):
    """
    Use this function as sorting key to sort in 'natural' order.

    >>> names = ['file10', 'file1.5', 'file1']
    >>> sorted(names, key=natural_sort_key)
    ['file1', 'file1.5', 'file10']

    Source: http://stackoverflow.com/a/16090640/715090
    """
    return [int(text) if text.isdigit() else text.lower()
        for text in re.split(_nsre, s)]


class UniqueNamer:
    """
    Assign unique names by appending letters to already seen names.
    """
    def __init__(self):
        self._names = set()

    def __call__(self, name):
        ext = 'A'
        new_name = name
        while new_name in self._names:
            if ext == '[':
                raise ValueError('Too many duplicate names')
            new_name = name + ext
            ext = chr(ord(ext) + 1)
        self._names.add(new_name)
        return new_name


class Merger:
    """
    Merge similar items into one. To specify what "similar" means, implement
    the merged() method in a subclass.
    """
    def __init__(self):
        self._items = []

    def add(self, item):
        # This method could possibly be made simpler if the graph structure
        # was made explicit.

        items = []
        for existing_item in self._items:
            m = self.merged(existing_item, item)
            if m is None:
                items.append(existing_item)
            else:
                item = m
        items.append(item)
        self._items = items

    def extend(self, iterable):
        for i in iterable:
            self.add(i)

    def __iter__(self):
        if self._items and hasattr(self._items, 'name'):
            yield from sorted(self._items, key=lambda x: x.name)
        else:
            yield from self._items

    def __len__(self):
        return len(self._items)

    def merged(self, existing_item, item):
        """
        If existing_item and item can be returned, this method must return
        a new item that represents a merged version of both. If they cannot
        be merged, it must return None.
        """
        raise NotImplementedError("not implemented")


def relative_symlink(src, dst, force=False):
    """
    Create a symbolic link in any directory.

    force -- if True, then overwrite an existing file/symlink
    """
    if force:
        try:
            os.remove(dst)
        except FileNotFoundError:
            pass
    target = os.path.relpath(os.path.abspath(src), start=os.path.dirname(dst))
    os.symlink(target, dst)


def nt_to_aa(s, _get=GENETIC_CODE.get):
    """Translate a nucleotide sequence to an amino acid sequence"""
    return ''.join(_get(s[i:i+3], '*') for i in range(0, len(s), 3))


def has_stop(sequence):
    """
    Return a boolean indicating whether the sequence has an internal stop codon.
    An incomplete codon at the end is allowed.

    >>> has_stop('GGG')
    False
    >>> has_stop('TAA')
    True
    >>> has_stop('GGGAC')
    False
    """
    s = sequence[:len(sequence) // 3 * 3]
    return '*' in nt_to_aa(s)


def plural_s(n):
    return 's' if n != 1 else ''


class FastaValidationError(Exception):
    pass


def validate_fasta(path):
    """
    Ensure that the FASTA file is suitable for use with makeblastdb.
    Raise a FastaValidationError if any of the following are true:

    - a record is empty
    - a record name occurs more than once
    - a sequence occurs more than once
    """
    with dnaio.open(path) as sr:
        records = list(sr)

    names = set()
    sequences = dict()
    for r in records:
        if len(r.sequence) == 0:
            raise FastaValidationError("Record {!r} is empty".format(r.name))
        if r.name in names:
            raise FastaValidationError("Record name {!r} occurs more than once".format(r.name))
        s = r.sequence.upper()
        if s in sequences:
            raise FastaValidationError("Records {!r} and {!r} contain the same sequence".format(
                r.name, sequences[s]))
        sequences[s] = r.name
        names.add(r.name)


def find_overlap(s, t, min_overlap=1):
    """
    Detect if s and t overlap.

    Returns:

    None if no overlap was detected.
    0 if s is a prefix of t or t is a prefix of s.
    Positive int gives index where t starts within s.
    Negative int gives -index where s starts within t.

    >>> find_overlap('ABCDE', 'CDE')
    2
    >>> find_overlap('CDE', 'ABCDEFG')
    -2
    >>> find_overlap('ABC', 'X') is None
    True
    """
    aligner = Aligner(s, max_error_rate=0, min_overlap=min_overlap)
    result = aligner.locate(t)
    if result is None:
        return None
    s_start, _, t_start, _, _, _ = result
    return s_start - t_start


def merge_overlapping(s, t):
    """
    Return merged sequences or None if they do not overlap
    The minimum overlap is 50% of the length of the shorter sequence.
    """
    i = find_overlap(s, t, min_overlap=max(1, min(len(s), len(t)) // 2))
    if i is None:
        return None

    if i >= 0:
        # positive: index of t in s
        if i + len(t) < len(s):
            # t is in s
            return s
        return s[:i] + t
    if -i + len(s) < len(t):
        # s is in t
        return t
    return t[:-i] + s


def get_cpu_time():
    """Return CPU time used by process and children"""
    if sys.platform != 'linux':
        return None
    rs = resource.getrusage(resource.RUSAGE_SELF)
    rc = resource.getrusage(resource.RUSAGE_CHILDREN)
    return rs.ru_utime + rs.ru_stime + rc.ru_utime + rc.ru_stime


def slice_arg(s):
    """
    Parse a string that describes a slice with start and end.

    >>> slice_arg('2:-3')
    slice(2, -3, None)

    >> slice_arg(':-3')
    slice(None, -3, None)

    >> slice_arg('2:')
    slice(2, None, None)
    """

    start, end = s.split(':')
    start = None if start == '' else int(start)
    end = None if end == '' else int(end)
    return slice(start, end)


def is_same_gene(name1: str, name2: str):
    """
    Compare gene names to find out whether they are alleles of each other.
    Both names must have a '*' in them
    """
    return '*' in name1 and '*' in name2 and name1.split('*')[0] == name2.split('*')[0]


class ChimeraFinder:
    def __init__(self, sequences: List[str], min_length: int=10):
        self._sequences = sequences
        self._min_length = min_length
        self._build_index()

    def _build_index(self):
        min_length = self._min_length
        # Create two dictionaries that map all prefixes and suffixes to indexes of all
        # sequences they occur in
        prefixes = defaultdict(list)
        suffixes = defaultdict(list)

        for i, sequence in enumerate(self._sequences):
            if len(sequence) < min_length:
                continue
            for stop in range(min_length, len(sequence) + 1):
                prefix = sequence[:stop]
                prefixes[prefix].append(i)
            for start in range(0, len(sequence) + 1 - min_length):
                suffix = sequence[start:]
                suffixes[suffix].append(i)
        self._prefixes = prefixes
        self._suffixes = suffixes

    def find_exact(self, query: str):
        """
        Find out whether the query string can be explained as a concatenation of
        a prefix of one of the strings plus a suffix of one of the strings in
        the given list of sequences. Both the prefix and the suffix must have a
        length of at least min_length.

        If the answer is yes, a tuple (prefix_length, prefix_indices, suffix_indices)
        is returned where prefix_length is the length of the query prefix,
        prefix_indices is a list of int indices into the sequences of all possible
        prefix sequences that match, and suffix_indices is the same for the suffix.

        The prefix_length returned is the first that yields a result. More are
        possible.

        If the answer is no, None is returned.
        """
        min_length = self._min_length
        for split_index in range(min_length, len(query) + 1 - min_length):
            prefix = query[:split_index]
            suffix = query[split_index:]
            if prefix in self._prefixes and suffix in self._suffixes:
                return (split_index, self._prefixes[prefix], self._suffixes[suffix])

        return None


def available_cpu_count():
    """
    Number of available virtual or physical CPUs on this system
    """
    if sys.platform != 'linux':
        try:
            import multiprocessing
            return multiprocessing.cpu_count()
        except (ImportError, NotImplementedError):
            return 1
    return len(os.sched_getaffinity(0))



"""
Species-specific code, such as lists of motifs and regular expressions.

Some refactoring is needed to make this module actually usable for many
species. Right now, it works for - at least - human, rhesus monkey and mouse.
"""
import re
from .utils import nt_to_aa


# Regular expressions for CDR3 detection
#
# The idea comes from D’Angelo et al.: The antibody mining toolbox.
# http://dx.doi.org/10.4161/mabs.27105
# The heavy-chain regex was taken directly from there, but the difference
# is that we express everything in terms of amino acids, not nucleotides.
# This simplifies the expressions and makes them more readable.
#
_CDR3_REGEX = {
    # Heavy chain
    'VH': re.compile("""
        [FY] [FHVWY] C
        (?P<cdr3>
            [ADEGIKMNRSTV] .{3,31}
        )
        W[GAV]
        """, re.VERBOSE),

    # Light chain, kappa
    'VK': re.compile("""
        [FSVY] [CFHNVY] [CDFGLSW]
        (?P<cdr3>
            .{4,15}
        )
        [FLV][GRV]
        """, re.VERBOSE),

    # Light chain, lambda
    'VL': re.compile("""
        # the negative lookahead assertion ensures that the rightmost start is found
        [CDY](?![CDY][CFHSY][CFGW])[CFHSY][CFGW]
        (?P<cdr3>
            .{4,15}
        )
        [FS]G
        """, re.VERBOSE)
}

_CDR3_VH_ALTERNATIVE_REGEX = re.compile("""
        C
        (?P<cdr3> . [RK] .{3,30})
        [WF]G.G
""", re.VERBOSE)


def find_cdr3(sequence, chain):
    """
    Find the CDR3 in the given sequence, assuming it comes from the given chain ('VH', 'VK', 'VL').
    If the chain is not one of 'VH', 'VK', 'VL', return None.

    Return a tuple (start, stop) if found, None otherwise.
    """
    try:
        regex = _CDR3_REGEX[chain]
    except KeyError:
        return None
    matches = []
    for offset in 0, 1, 2:
        aa = nt_to_aa(sequence[offset:])
        match = regex.search(aa)
        if not match and chain == 'VH':
            match = _CDR3_VH_ALTERNATIVE_REGEX.search(aa)
        if match:
            start, stop = match.span('cdr3')
            matches.append((start * 3 + offset, stop * 3 + offset))
    return min(matches, default=None)


# The following code is used for detecting CDR3 start sites within V
# reference sequences and CDR3 end sites within J reference sequences.


# Matches the start of the CDR3 within the end of a VH sequence
_CDR3_START_VH_REGEX = re.compile("""
    [FY] [FHVWY] C
    (?P<cdr3_start>
        [ADEGIKMNRSTV*] | $
    )
    """, re.VERBOSE)


_CDR3_START_VH_ALTERNATIVE_REGEX = re.compile("""
    C
    (?P<cdr3_start> . [RK])
    """, re.VERBOSE)


_CDR3_START_REGEXES = {
    'kappa': re.compile('[FSVY][CFHNVY][CDFGLSW]'),
    'lambda': re.compile('[CDY](?![CDY][CFHSY][CFGW])[CFHSY][CFGW]'),
    'gamma': re.compile('[YFH]C'),  # TODO test whether this also works for alpha and beta
    'delta': re.compile('[YFH]C'),
}


def _cdr3_start_heavy(aa):
    head, tail = aa[:-15], aa[-15:]
    match = _CDR3_START_VH_REGEX.search(tail)
    if not match:
        match = _CDR3_START_VH_ALTERNATIVE_REGEX.search(tail)
    if not match:
        return None
    return len(head) + match.start('cdr3_start')


def cdr3_start(nt, chain):
    """
    Find CDR3 start location within a V gene (Ig or TCR)

    nt -- nucleotide sequence of the gene
    chain -- one of the following strings:
      - 'heavy', 'lambda', 'kappa' for Ig genes
      - 'alpha', 'beta', 'gamma', 'delta' for TCR genes
    """
    aa = nt_to_aa(nt)
    if chain == 'heavy':
        start = _cdr3_start_heavy(aa)
        if start is None:
            return None
        return 3 * start
    if chain in ('kappa', 'lambda', 'gamma', 'delta'):
        head, tail = aa[:-15], aa[-15:]
        match = _CDR3_START_REGEXES[chain].search(tail)
        if match:
            return 3 * (len(head) + match.end())
        else:
            return None
    elif chain in ('alpha', 'beta'):
        head, tail = aa[:-8], aa[-8:]
        pos = tail.find('C')
        if pos == -1:
            return None
        else:
            return 3 * (len(head) + pos + 1)


# Matches after the end of the CDR3 within a J sequence
_CDR3_END_REGEXES = {
    'heavy': re.compile('W[GAV]'),
    'kappa': re.compile('FG'),
    'lambda': re.compile('FG'),
    'alpha': re.compile('FG'),
    'beta': re.compile('FG'),
    'gamma': re.compile('FG'),
    'delta': re.compile('FG'),
}


def cdr3_end(nt, chain):
    """
    Find the position of the CDR3 end within a J sequence

    nt -- nucleotide sequence of the J gene
    chain -- one of the following strings:
      - 'heavy', 'lambda', 'kappa' for Ig genes
      - 'alpha', 'beta', 'gamma', 'delta' for TCR genes
    """
    regex = _CDR3_END_REGEXES[chain]
    for frame in 0, 1, 2:
        aa = nt_to_aa(nt[frame:])
        match = regex.search(aa)
        if match:
            return match.start() * 3 + frame
    return None


# When searching for the CDR3, start this many bases to the left of the end of
# the V match.
CDR3_SEARCH_START = 30



#!/usr/bin/env python3
"""
IgDiscover computes V/D/J gene usage profiles and discovers novel V genes

- Run IgBLAST in parallel (wrapper inspired by igblastwrp).
- Parse IgBLAST output into a tab-separated table
- Group sequences by barcode
- Plot V gene usage
- Discover new V genes given more than one dataset
"""
import sys
import logging
import pkgutil
import importlib
from argparse import ArgumentParser, RawDescriptionHelpFormatter
import matplotlib as mpl
import warnings
import resource

import igdiscover.cli as cli_package

from . import __version__

__author__ = "Marcel Martin"

mpl.use('Agg')
warnings.filterwarnings('ignore', 'axes.color_cycle is deprecated and replaced with axes.prop_cycle')
warnings.filterwarnings('ignore', 'The `IPython.html` package')
warnings.filterwarnings('ignore', 'Widget registration using a string name has been deprecated')
warnings.filterwarnings('ignore', 'Traits should be given as instances, not types')

logger = logging.getLogger(__name__)


class HelpfulArgumentParser(ArgumentParser):
    """An ArgumentParser that prints full help on errors."""

    def __init__(self, *args, **kwargs):
        if 'formatter_class' not in kwargs:
            kwargs['formatter_class'] = RawDescriptionHelpFormatter
        super().__init__(*args, **kwargs)

    def error(self, message):
        self.print_help(sys.stderr)
        args = {'prog': self.prog, 'message': message}
        self.exit(2, '%(prog)s: error: %(message)s\n' % args)


def format_duration(seconds):
    h = int(seconds // 3600)
    seconds -= h * 3600
    m = int(seconds // 60)
    seconds -= m * 60
    return '{:02d}:{:02d}:{:04.1f}'.format(h, m, seconds)


def main(arguments=None):
    logging.basicConfig(level=logging.INFO, format='%(levelname)s: %(message)s')
    parser = HelpfulArgumentParser(description=__doc__, prog='igdiscover')
    parser.add_argument('--profile', default=False, action='store_true',
        help='Save profiling information to igdiscover.prof')
    parser.add_argument('--version', action='version', version='%(prog)s ' + __version__)

    show_cpustats = dict()
    subparsers = parser.add_subparsers()

    # Import each module that implements a subcommand and add a subparser for it.
    # Each subcommand is implemented as a module in the cli subpackage.
    # It needs to implement an add_arguments() and a main() function.
    modules = pkgutil.iter_modules(cli_package.__path__)
    for _, module_name, _ in modules:
        module = importlib.import_module("." + module_name, cli_package.__name__)
        subparser = subparsers.add_parser(module_name,
            help=module.__doc__.split('\n')[1], description=module.__doc__)
        subparser.set_defaults(func=module.main)
        module.add_arguments(subparser)
        if hasattr(module, 'do_not_show_cpustats'):
            show_cpustats[module.main] = False

    args = parser.parse_args(arguments)
    if not hasattr(args, 'func'):
        parser.error('Please provide the name of a subcommand to run')
    elif args.profile:
        import cProfile as profile
        profile.runctx('args.func(args)', globals(), locals(), filename='igdiscover.prof')
        logger.info('Wrote profiling data to igdiscover.prof')
    else:
        args.func(args)
    if sys.platform == 'linux' and show_cpustats.get(args.func, True):
        rself = resource.getrusage(resource.RUSAGE_SELF)
        rchildren = resource.getrusage(resource.RUSAGE_CHILDREN)
        memory_kb = rself.ru_maxrss + rchildren.ru_maxrss
        cpu_time = rself.ru_utime + rself.ru_stime + rchildren.ru_utime + rchildren.ru_stime
        cpu_time_s = format_duration(cpu_time)
        logger.info('CPU time {}. Maximum memory usage {:.3f} GB'.format(
            cpu_time_s, memory_kb / 1E6))


if __name__ == '__main__':
    main()



"""
Test the SiblingMerger class
"""
import pandas as pd
import pytest

from igdiscover.cli.discover import SiblingMerger, SiblingInfo
from igdiscover.utils import UniqueNamer
from igdiscover.cli.rename import PrefixDict


def test_0():
    merger = SiblingMerger()
    assert list(merger) == []


def test_1():
    merger = SiblingMerger()
    group = pd.DataFrame([1, 2, 3])
    info = SiblingInfo('ACCGGT', False, 'name1', group)
    merger.add(info)
    assert list(merger) == [info]


def test_2():
    merger = SiblingMerger()
    group = pd.DataFrame([1, 2, 3])
    merger.add(SiblingInfo('ACCGGT', False, 'name1', group))
    merger.add(SiblingInfo('ACCGGT', False, 'name2', group))
    sisters = list(merger)
    assert len(sisters) == 1
    assert sisters[0].sequence == 'ACCGGT'
    assert not sisters[0].requested
    assert sisters[0].name == 'name1;name2'


def test_requested():
    merger = SiblingMerger()
    group = pd.DataFrame([1, 2, 3])
    merger.add(SiblingInfo('ACCGGT', True, 'name1', group))
    merger.add(SiblingInfo('ACCGGT', False, 'name2', group))
    sisters = list(merger)
    assert sisters[0].requested


def test_prefix():
    merger = SiblingMerger()
    group = pd.DataFrame([1, 2, 3])
    merger.add(SiblingInfo('ACCGGTAACGT', True, 'name1', group))
    merger.add(SiblingInfo('ACCGGT', False, 'name2', group))
    info2 = SiblingInfo('TGATACC', False, 'name3', group)
    merger.add(info2)
    sisters = list(merger)
    assert len(sisters) == 2
    assert sisters[0].sequence == 'ACCGGTAACGT'
    assert sisters[0].name == 'name1;name2'
    assert sisters[1] == info2


def test_with_N():
    merger = SiblingMerger()
    group = pd.DataFrame([1, 2, 3])
    merger.add(SiblingInfo('ACCNGTAANGT', True, 'name1', group))
    merger.add(SiblingInfo('ANCGGT', False, 'name2', group))
    info2 = SiblingInfo('TGATACC', False, 'name3', group)
    merger.add(info2)
    sisters = list(merger)
    assert len(sisters) == 2
    assert sisters[0].sequence == 'ACCGGTAANGT'
    assert sisters[0].name == 'name1;name2'
    assert sisters[1] == info2


def test_unique_namer():
    namer = UniqueNamer()
    assert namer('Name') == 'Name'
    assert namer('AnotherName') == 'AnotherName'
    assert namer('Name') == 'NameA'
    assert namer('Name') == 'NameB'
    assert namer('YetAnotherName') == 'YetAnotherName'
    assert namer('Name') == 'NameC'
    assert namer('NameC') == 'NameCA'


class TestPrefixDict:
    def setup(self):
        self.pd = PrefixDict([
            ('AAACCT', 7),
            ('AGAAA', 11),
            ('AGAAC', 13)
        ])

    def test_ambiguous(self):
        with pytest.raises(KeyError):
            self.pd['AGAA']

    def test_missing(self):
        with pytest.raises(KeyError):
            self.pd['TTAAG']

    def test_existing(self):
        assert self.pd['AAACCT'] == 7
        assert self.pd['AAACCT'] == 7
        assert self.pd['AAA'] == 7
        assert self.pd['AGAAAT'] == 11
        assert self.pd['AGAACT'] == 13



from collections import OrderedDict, defaultdict
import pandas as pd
from scipy.spatial import distance
from scipy.cluster import hierarchy
from tinyalign import hamming_distance
from .utils import distances
from .trie import Trie


def all_nodes(root):
    """Return a list of all nodes of the tree, from left to right (iterative implementation)."""
    result = []
    path = [None]
    node = root
    while node is not None:
        if node.left is not None:
            path.append(node)
            node = node.left
        elif node.right is not None:
            result.append(node)
            node = node.right
        else:
            result.append(node)
            node = path.pop()
            if node is not None:
                result.append(node)
                node = node.right

    return result


def inner_nodes(root):
    """
    Return a list of all inner nodes of the tree, from left to right.
    """
    return [node for node in all_nodes(root) if not node.is_leaf()]


def collect_ids(root):
    """
    Return a list of ids of all leaves of the given tree
    """
    return [node.id for node in all_nodes(root) if node.is_leaf()]


def cluster_sequences(sequences, minsize=5):
    """
    Cluster the given sequences into groups of similar sequences.

    Return a triple that contains a pandas.DataFrame with the edit distances,
    the linkage result, and a list that maps sequence ids to their cluster id.
    If an entry is zero in that list, it means that the sequence is not part of
    a cluster.
    """
    matrix = distances(sequences)
    linkage = hierarchy.linkage(distance.squareform(matrix), method='average')
    # Linkage columns are:
    # 0, 1: merged clusters, 2: distance, 3: number of nodes in cluster
    inner = inner_nodes(hierarchy.to_tree(linkage))
    prev = linkage[:, 2].max()  # highest distance
    clusters = [0] * len(sequences)
    cl = 1
    for n in inner:
        if n.dist > 0 and prev / n.dist < 0.8 \
                and n.left.count >= minsize and n.right.count >= minsize:
            for id in collect_ids(n.left):
                # Do not overwrite previously assigned ids
                if clusters[id] == 0:
                    clusters[id] = cl
            cl += 1
        prev = n.dist
    # At the end of the above loop, we have not processed the rightmost
    # subtree. In our experiments, it never contains true novel sequences,
    # so we omit it.

    return pd.DataFrame(matrix), linkage, clusters


class Graph:
    """Graph that can find connected components"""
    def __init__(self, nodes):
        self._nodes = OrderedDict()
        for node in nodes:
            self._nodes[node] = []

    def add_edge(self, node1, node2):
        self._nodes[node1].append(node2)
        self._nodes[node2].append(node1)

    def connected_components(self):
        """Return a list of connected components."""
        visited = set()
        components = []
        for node, neighbors in self._nodes.items():
            if node in visited:
                continue
            # Start a new component
            to_visit = [node]
            component = []
            while to_visit:
                n = to_visit.pop()
                if n in visited:
                    continue
                visited.add(n)
                component.append(n)
                for neighbor in self._nodes[n]:
                    if neighbor not in visited:
                        to_visit.append(neighbor)
            components.append(component)
        return components


def cluster_by_length(strings):
    """
    Cluster a set of strings by length
    """
    string_lists = defaultdict(list)
    for s in strings:
        string_lists[len(s)].append(s)
    return list(string_lists.values())


def single_linkage(strings, linked):
    """
    Cluster a set of strings. *linked* is a function with two parameters s and t that returns
    whether *s* and *t* are in the same cluster.

    >>> single_linkage(['ABC', 'ABD', 'DEFG', 'DEFH'], lambda s, t: len(s) == len(t) and
    ...        hamming_distance(s, t) <= 1)
    [['ABC', 'ABD'], ['DEFG', 'DEFH']]
    """
    graph = Graph(strings)
    for i, s in enumerate(strings):
        for j, t in enumerate(strings[i + 1:]):
            if linked(s, t):
                graph.add_edge(s, t)
    return graph.connected_components()


def hamming_single_linkage(strings, mismatches, linked=None):
    """
    Cluster a set of strings by hamming distance. Strings with
    a distance of at most 'mismatches' will be put into the same cluster.

    Uses the optimization that strings of different lengths can be
    clustered separately.

    Use *linked* to override the function passed to single_linkage. It will
    only be called for strings of the same length.

    Return a list of connected components (clusters).
    """
    if linked is None:
        def linked(s, t):
            return hamming_distance(s, t) <= mismatches

    components = []
    for strings in cluster_by_length(strings):
        components.extend(single_linkage(strings, linked))
    return components


def hamming_single_linkage_trie(strings, mismatches):
    """
    Cluster by hamming distance using a trie
    """
    components = []
    for strings in cluster_by_length(strings):
        graph = Graph(strings)
        trie = Trie()
        for s in strings:
            trie.add(s)
        for s in strings:
            for neighbor in trie.find_all_similar(s, mismatches):
                if neighbor != s:
                    graph.add_edge(s, neighbor)
        components.extend(graph.connected_components())
    return components



"""
Group sequences that share a unique molecular identifier (UMI, barcode)

Since the same barcode can sometimes be used by different sequences, the CDR3
sequence can further be used to distinguish sequences. You can choose between
using either a 'pseudo CDR3' sequence, which encompasses by default bases 80
to 61 counted from the 3' end. Or you can use the real CDR3 detected with a
regular expression.

If grouping by CDR3s is enabled, sequences with identical barcode and CDR3
must additionally have a similar length. If the length differs by more than
2 bp, they are put into different groups.

The barcode can be in the 5' end or the 3' end of the sequence.

Use --trim-g to remove initial runs of G at the 5' end (artifact from RACE protocol).
These are removed after the barcode is removed.

For all the found groups, one sequence is output to standard output (in FASTA
format). Which sequence that is depends on the group size:
- If the group consists of a single sequence, that sequence is output
- If the group consists of two sequences, one of them is picked randomly
- If the group has at least three sequences, a consensus is computed. The
  consensus is output if it contains no ambiguous bases. Otherwise, also here a
  random sequence is chosen.
"""
# NOTES
#
# - Different lengths of the initial G run cannot be used to distinguish sequences
#  since they can come from polymerase problems in homopolymers.
# - There are some indels in homopolymers (probably PCR problem)
# - There are also regular sequencing errors in the initial run of G nucleotides.
# - Some paired reads aren’t correctly merged into single reads. They end up being
#   too long.
# - When grouping by barcode and pseudo CDR3, sequence lengths vary within groups.
#   However, this affects only ~1% of sequences, so it is not necessary to compute
#   a multiple alignment. Just taking the consensus will drown the incorrect
#   sequences, at least if the group size is large.
# - It does not hurt to reduce the minimimum number of sequences per group for
#   taking a consensus to 2, but it also does not help much (results in 0.5% more
#   sequences). (The consensus is only successful if both sequences are identical.)
#   However, since this is also a simple way to deal with exact duplicates, we do
#   it anyway and can then skip the separate duplicate removal step (VSEARCH).

# TODO
# - Use pandas.DataFrame

import csv
import sys
import logging
from collections import Counter, defaultdict
from contextlib import ExitStack
from itertools import islice
import json
from typing import List, Iterable

import dnaio
from xopen import xopen

from ..align import consensus
from ..species import find_cdr3
from ..cluster import Graph
from ..utils import slice_arg

# minimum number of sequences needed for attempting to compute a consensus
MIN_CONSENSUS_SEQUENCES = 3


logger = logging.getLogger(__name__)


def add_arguments(parser):
    arg = parser.add_argument
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--real-cdr3', action='store_true', default=False,
        help='In addition to barcode, group sequences by real CDR3 (detected with regex).')
    group.add_argument('--pseudo-cdr3', nargs='?', default=None,
        type=slice_arg, const=slice(-80, -60), metavar='START:END',
        help='In addition to barcode, group sequences by pseudo CDR3. '
            'If START:END is omitted, use -80:-60.')
    arg('--groups-output', metavar='FILE', default=None,
        help='Write tab-separated table with groups to FILE')
    arg('--plot-sizes', metavar='FILE', default=None,
        help='Plot group sizes to FILE (.png or .pdf)')
    arg('--limit', default=None, type=int, metavar='N',
        help='Limit processing to the first N reads')
    arg('--trim-g', action='store_true', default=False,
        help="Trim 'G' nucleotides at 5' end")
    arg('--minimum-length', '-l', type=int, default=0,
        help='Minimum sequence length')
    arg('--barcode-length', '-b', type=int, default=12,
        help="Length of barcode. Positive for 5' barcode, negative for 3' barcode. Default: %(default)s")
    arg('--json', metavar="FILE", help="Write statistics to FILE")
    arg('fastx', metavar='FASTA/FASTQ',
        help='FASTA or FASTQ file (can be gzip-compressed) with sequences')


def hamming_neighbors(s: str) -> Iterable[str]:
    """Return sequences that are at hamming distance 1 and return also s itself"""
    for i in range(len(s)):
        for c in 'ACGT':
            if s[i] != c:
                yield s[:i] + c + s[i+1:]
    yield s


def cluster_by_cdr3(records, pseudo_cdr3: slice, real_cdr3: bool) -> List[List]:
    """
    Single-linkage clustering of sequences based on sequence length and CDR3 similarity.

    Two sequences are linked if
    - their (pseudo-) CDR3 sequences have a hamming distance of at most 1
    - and their full lengths (not the CDR3 length) differs by at most 2.

    Sequences for which no CDR3 could be detected are filtered out.
    """
    assert pseudo_cdr3 or real_cdr3

    records_cdr3s = []
    for record in records:
        if real_cdr3:
            match = find_cdr3(record.sequence, chain='VH')
            if match:
                cdr3 = record.sequence[match[0]:match[1]]
            else:
                continue
        else:
            assert pseudo_cdr3
            cdr3 = record.sequence[pseudo_cdr3]
        records_cdr3s.append((record, cdr3))
    del records

    # logger.info('%s times (%.2f%%), the CDR3 regex matched', n, n / (n + regex_fail) * 100)

    if len(records_cdr3s) == 1:  # TODO check if this helps
        return [records_cdr3s]

    cdr3s = [rc[1] for rc in records_cdr3s]

    # Cluster by CDR3 sequences
    unique_cdr3s = set(cdr3s)
    sorted_cdr3s = sorted(unique_cdr3s)  # For reproducibility
    graph = Graph(sorted_cdr3s)
    for cdr3 in sorted_cdr3s:
        for neighbor in hamming_neighbors(cdr3):
            if neighbor in unique_cdr3s:
                graph.add_edge(cdr3, neighbor)
    cdr3_clusters = graph.connected_components()

    # Map each CDR3 sequence to a list of records with that CDR3
    cdr3_records = defaultdict(list)
    for record, cdr3 in records_cdr3s:
        cdr3_records[cdr3].append(record)

    # To compute the final clusters that also take length into account,
    # split the CDR3 clusters where necessary
    clusters = []
    for cdr3_cluster in cdr3_clusters:
        # Gather all records in this CDR3 cluster
        cluster_records = []
        for cdr3 in cdr3_cluster:
            for record in cdr3_records[cdr3]:
                cluster_records.append((record, cdr3))

        # For "clustering" by sequence length, sort by sequence length
        # and look at adjacent records: If their length differs too much,
        # the cluster needs to be split
        cluster_records.sort(key=lambda rc: len(rc[0].sequence))
        cluster = []
        prev_length = None
        for record, cdr3 in cluster_records:
            l = len(record.sequence)
            if prev_length is not None and l > prev_length + 2:
                # Start a new component
                clusters.append(cluster)
                cluster = []
            cluster.append((record, cdr3))
            prev_length = l
        if cluster:
            clusters.append(cluster)

    assert sum(len(component) for component in clusters) == len(records_cdr3s)
    assert all(clusters)  # Components must be non-empty
    return clusters


GROUPS_HEADER = ['barcode', 'cdr3', 'name', 'sequence']


def write_group(csvfile, barcode, sequences, write_cdr3: bool):
    for sequence, cdr3 in sequences:
        row = [barcode, sequence.name.split(maxsplit=1)[0], sequence.sequence]
        if write_cdr3:
            row[1:1] = [cdr3]
        csvfile.writerow(row)
    csvfile.writerow([])


def collect_barcode_groups(fastx, barcode_length, trim_g, limit, minimum_length):
    """
    fastx -- path to FASTA or FASTQ input

    """
    cdr3s = set()
    # Map barcodes to lists of sequences
    barcodes = defaultdict(list)
    n = 0
    too_short = 0
    with dnaio.open(fastx) as f:
        for record in islice(f, 0, limit):
            if len(record) < minimum_length:
                too_short += 1
                continue
            if barcode_length > 0:
                barcode = record.sequence[:barcode_length]
                unbarcoded = record[barcode_length:]
            else:
                barcode = record.sequence[barcode_length:]
                unbarcoded = record[:barcode_length]

            if trim_g:
                # The RACE protocol leads to a run of non-template Gs in the beginning
                # of the sequence, after the barcode.
                unbarcoded.sequence = unbarcoded.sequence.lstrip('G')
                if unbarcoded.qualities:
                    unbarcoded.qualities = unbarcoded.qualities[-len(unbarcoded.sequence):]

            barcodes[barcode].append(unbarcoded)
            n += 1

    logger.info('%s sequences in input', n + too_short)
    logger.info('%s sequences long enough', n)

    return barcodes


def plot_sizes(sizes, path):
    from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
    from matplotlib.figure import Figure
    import matplotlib
    import seaborn as sns
    sns.set()
    fig = Figure()
    matplotlib.rcParams.update({'font.size': 14})
    FigureCanvas(fig)
    ax = fig.add_subplot(111)
    v, _, _ = ax.hist(sizes, bins=100)
    ax.set_ylim(0, v[1:].max() * 1.1)
    ax.set_xlabel('Group size')
    ax.set_ylabel('Read frequency')
    ax.set_title('Histogram of group sizes (>1)')
    ax.grid(axis='x')
    ax.tick_params(direction='out', top=False, right=False)
    fig.set_tight_layout(True)
    fig.savefig(path)
    logger.info('Plotted group sizes to %r', path)


def main(args):
    if args.barcode_length == 0:
        sys.exit("Barcode length must be non-zero")

    barcodes = collect_barcode_groups(args.fastx, args.barcode_length, args.trim_g,
        args.limit, args.minimum_length)

    logger.info('%s unique barcodes', len(barcodes))
    barcode_singletons = sum(1 for seqs in barcodes.values() if len(seqs) == 1)
    logger.info('%s barcodes used by only a single sequence (singletons)', barcode_singletons)

    group_by_cdr3 = args.pseudo_cdr3 or args.real_cdr3

    with ExitStack() as stack:
        if args.groups_output:
            group_out = csv.writer(stack.enter_context(
                xopen(args.groups_output, 'w')), delimiter='\t', lineterminator='\n')
            group_out.writerow(GROUPS_HEADER)
        else:
            group_out = None
        too_few = 0
        n_clusters = 0
        n_singletons = 0
        n_consensus = 0
        n_ambiguous = 0
        sizes = []
        for barcode in sorted(barcodes):
            sequences = barcodes[barcode]
            if len(sequences) != len(set(sequence.name for sequence in sequences)):
                logger.error('Duplicate sequence records detected')
                sys.exit(1)
            if group_by_cdr3:
                clusters = cluster_by_cdr3(sequences, args.pseudo_cdr3, args.real_cdr3)
            else:
                # TODO it would be useful to do the length clustering that cluster_by_cdr3() does
                clusters = [[(record, None) for record in sequences]]
            n_clusters += len(clusters)
            for cluster in clusters:

                sizes.append(len(cluster))
                if group_out:
                    write_group(group_out, barcode, cluster, write_cdr3=group_by_cdr3)
                if len(cluster) == 1:
                    n_singletons += 1
                if len(cluster) < MIN_CONSENSUS_SEQUENCES:

                    sequence = cluster[0][0].sequence
                    name = cluster[0][0].name
                    cdr3 = cluster[0][1]
                    too_few += 1
                else:
                    cons = consensus({record.name: record.sequence for record, cdr3 in cluster}, threshold=0.501)
                    if 'N' in cons:
                        # Pick the first sequence as the output sequence
                        sequence = cluster[0][0].sequence
                        name = cluster[0][0].name
                        cdr3 = cluster[0][1]
                        n_ambiguous += 1
                    else:
                        sequence = cons
                        n_consensus += 1
                        name = 'consensus{}'.format(n_consensus)
                        cdr3 = Counter(cdr3 for record, cdr3 in cluster).most_common(1)[0][0]

                name = name.split(maxsplit=1)[0]
                if name.endswith(';'):
                    name = name[:-1]
                if group_by_cdr3:
                    print('>{};barcode={};cdr3={};size={};\n{}'.format(name, barcode,
                        cdr3, len(cluster), sequence))
                else:
                    print('>{};barcode={};size={};\n{}'.format(name, barcode,
                        len(cluster), sequence))

    logger.info('%d clusters (%d singletons)', n_clusters, n_singletons)
    logger.info('%d consensus sequences computed (from groups that had at least %d sequences)',
        n_consensus + n_ambiguous, MIN_CONSENSUS_SEQUENCES)
    logger.info('%d of those had no ambiguous bases', n_consensus)
    if args.groups_output:
        logger.info('Groups written to %r', args.groups_output)

    if args.json:
        sizes_counter = Counter(sizes)
        stats = {
            'unique_barcodes': len(barcodes),
            'barcode_singletons': barcode_singletons,
            'groups_written': n_clusters,
            'group_size_1': sizes_counter[1],
            'group_size_2': sizes_counter[2],
            'group_size_3plus': sum(v for k, v in sizes_counter.items() if k >= 3),
        }
        with open(args.json, 'w') as f:
            json.dump(stats, f, indent=2)
            print(file=f)

    if args.plot_sizes:
        plot_sizes(sizes, args.plot_sizes)



import subprocess
from collections import Counter, OrderedDict
from itertools import groupby
from io import BytesIO

import dnaio

from .utils import available_cpu_count, downsampled


def multialign(sequences, program="mafft", threads=available_cpu_count()):
    """
    Wrapper for multiple sequence alignment tools. Currently supported are
    * ClustalO (http://www.clustal.org/omega/),
    * MAFFT (http://mafft.cbrc.jp/alignment/software/)
    * MUSCLE (http://www.drive5.com/muscle/)

    A package using the libclustalo library directly also exists:
    https://github.com/benchling/clustalo-python/blob/master/clustalo.c
    It is not for Python 3.

    sequences -- a dictionary mapping names to sequences.
        Use an OrderedDict if order matters.

    program -- must be "clustalo", "mafft", "muscle", "muscle-medium",
        "muscle-fast". The latter calls MUSCLE with parameters that make it run
        faster (but less accurate). "muscle-medium" is in between muscle and
        muscle-fast.

    threads -- number of threads to use for those programs that support it.
        By default, set to the number of processors.
    """
    if program == "mafft":
        args = ["mafft", "--quiet", "--thread", str(threads), "-"]
    elif program == "clustalo":
        args = ["clustalo", "--threads=" + str(threads), "--infile=-"]
    elif program == "muscle":
        args = ["muscle", "-quiet", "-in", "-", "-out", "-"]
    elif program == "muscle-fast":
        args = ["muscle", "-quiet", "-maxiters", "1", "-diags", "-in", "-", "-out", "-"]
    elif program == "muscle-medium":
        args = ["muscle", "-quiet", "-maxiters", "2", "-diags", "-in", "-", "-out", "-"]
    else:
        raise ValueError("program {!r} not supported".format(program))

    fasta_data = "".join(
        ">{}\n{}\n".format(name, seq) for name, seq in sequences.items()
    ).encode()
    result = subprocess.check_output(args, input=fasta_data)
    with dnaio.open(BytesIO(result)) as f:
        aligned = list(f)

    return {record.name: record.sequence.upper() for record in aligned}


def consensus(aligned, threshold=0.7, ambiguous='N', keep_gaps=False):
    """
    Compute a consensus from multialign() output, allowing degraded sequences
    in the 3' end.

    aligned -- a dict mapping names to sequences or a list of sequences
    keep_gaps -- whether the returned sequence contains gaps (-)
    """
    n = len(aligned)
    result = []
    if hasattr(aligned, 'values'):
        sequences = aligned.values()
    else:
        sequences = aligned

    active = int(len(aligned) * 0.05)
    for i, chars in enumerate(reversed(list(zip(*sequences)))):
        counter = Counter(chars)
        active = max(n - counter['-'], active)
        assert counter['-'] >= n - active
        counter['-'] -= n - active
        char, freq = counter.most_common(1)[0]
        if i >= 10:  # TODO hard-coded
            active = n
        if freq / active >= threshold:
            if keep_gaps or char != '-':
                result.append(char)
        else:
            result.append(ambiguous)
    return ''.join(result[::-1])


def iterative_consensus(sequences, program='muscle-medium', threshold=0.6,
        subsample_size=200, maximum_subsample_size=1600):
    """
    Compute a consensus sequence of the given sequences, but do not use all
    sequences if there are many: First, try to compute the consensus from a
    small subsample. If there are 'N' bases, increase the subsample size and
    repeat until either there are no more 'N' bases, all available sequences
    have been used or maximum_subsample_size is reached.
    """
    while True:
        sample = downsampled(sequences, subsample_size)
        aligned = multialign(OrderedDict(enumerate(sample)), program=program)

        cons = consensus(aligned, threshold=threshold).strip('N')
        if 'N' not in cons:
            # This consensus is good enough
            break
        if len(sequences) <= subsample_size:
            # We have already used all the sequences that are available
            break
        subsample_size *= 2
        if subsample_size > maximum_subsample_size:
            break
    return cons


def describe_nt_change(s: str, t: str):
    """
    Describe changes between two nucleotide sequences

    >>> describe_nt_change('AAA', 'AGA')
    '2A>G'
    >>> describe_nt_change('AAGG', 'AATTGG')
    '2_3insTT'
    >>> describe_nt_change('AATTGG', 'AAGG')
    '3_4delTT'
    >>> describe_nt_change('AATTGGCG', 'AAGGTG')
    '3_4delTT; 7C>T'
    >>> describe_nt_change('AAGCTAA', 'AACTGAA')
    '3G>C; 4C>T; 5T>G'
    """
    alignment = align_global(s, t, insertion=-4, deletion=-4)
    row1 = alignment.ref_row.replace('\0', '-')
    row2 = alignment.query_row.replace('\0', '-')
    changes = []

    def grouper(c):
        c1, c2 = c
        if c1 == c2:
            return 'MATCH'
        elif c1 == '-':
            return 'INS'
        elif c2 == '-':
            return 'DEL'
        else:
            return 'SUBST'

    index = 1
    for event, group in groupby(zip(row1, row2), grouper):
        if event == 'MATCH':
            index += len(list(group))
        elif event == 'SUBST':
            # ungroup
            for c1, c2 in group:
                change = '{}{}>{}'.format(index, c1, c2)
                changes.append(change)
                index += 1
        elif event == 'INS':
            inserted = ''.join(c[1] for c in group)
            change = '{}_{}ins{}'.format(index-1, index, inserted)
            changes.append(change)
        elif event == 'DEL':
            deleted = ''.join(c[0] for c in group)
            change = '{}_{}del{}'.format(index, index + len(deleted) - 1, deleted)
            changes.append(change)
            index += len(deleted)
    return '; '.join(changes)


class Alignment:
    def __init__(self, ref_row, query_row, ref_start, ref_stop, query_start, query_stop, score, errors):
        assert len(ref_row) == len(query_row)
        self.ref_row = ref_row
        self.ref_start = ref_start
        self.ref_stop = ref_stop
        self.query_row = query_row
        self.query_start = query_start
        self.query_stop = query_stop
        self.score = score
        self.errors = errors


def align_global_core(ref: str, query: str, match=1, mismatch=-2, insertion=-2, deletion=-2):
    """
    Compute an optimal global alignment between strings *ref* and *query*.
    An alignment is optimal if it has maximal score.

    Return an Alignment object.
    """
    m = len(ref)
    n = len(query)
    # DP Matrix:
    #            query (j)
    #          ----------> n
    #         |
    # ref (i) |
    #         |
    #         V
    #         m

    # the DP matrix is stored column-major
    scores = [[0] * (n + 1) for _ in range(m+1)]
    backtrace = [[0] * (n + 1) for _ in range(m+1)]

    start_in_ref = False
    start_in_query = False
    stop_in_ref = False
    stop_in_query = False

    UP = 0
    LEFT = 1
    DIAG = 2
    GAPCHAR = '\0'

    # initialize first column
    for i in range(m + 1):
        scores[i][0] = 0 if start_in_ref else i * deletion
        backtrace[i][0] = UP

    # initialize first row
    for j in range(n + 1):
        scores[0][j] = 0 if start_in_query else j * insertion
        backtrace[0][j] = LEFT

    # fill the entire DP matrix
    # outer loop goes over columns
    for j in range(1, n + 1):
        for i in range(1, m + 1):
            bt = DIAG
            score = scores[i-1][j-1] + (match if (ref[i-1] == query[j-1]) else mismatch)
            tmp = scores[i-1][j] + insertion
            if tmp > score:
                bt = UP
                score = tmp
            tmp = scores[i][j-1] + deletion
            if tmp > score:
                bt = LEFT
                score = tmp
            scores[i][j] = score
            backtrace[i][j] = bt

    # initialize best score and its position to the bottomright cell
    best_i = m
    best_j = n
    best_score = scores[m][n]

    if stop_in_query:
        # search also in last row
        for j in range(n + 1):
            if scores[m][j] >= best_score:
                best_score = scores[m][j]
                best_i = m
                best_j = j

    if stop_in_ref:
        # search also in last column
        for i in range(m + 1):
            if scores[i][n] >= best_score:
                best_i = i
                best_j = n
                best_score = scores[i][n]

    # trace back
    p1 = []
    p2 = []
    i = m
    j = n

    # first, walk from the lower right corner to the
    # position where we found the maximum score

    errors = 0

    # if gaps are currently errors, this is 1, otherwise it's 0
    gaps_are_errors = 0 if stop_in_query else 1
    if i == best_i:  # we are in the last row
        while j > best_j:
            p1.append(GAPCHAR)
            j -= 1
            p2.append(query[j])
            errors += gaps_are_errors
    else:  # we are in the last column
        gaps_are_errors = 0 if stop_in_ref else 1
        while i > best_i:
            i -= 1
            p1.append(ref[i])
            p2.append(GAPCHAR)
            errors += gaps_are_errors

    assert i == best_i and j == best_j

    # the actual backtracing
    # The alignments are constructed in reverse
    # and this is undone afterwards.
    while i > 0 and j > 0:
        direction = backtrace[i][j]
        if direction == DIAG:
            i -= 1
            j -= 1
            if ref[i] != query[j]:
                errors += 1
            p1.append(ref[i])
            p2.append(query[j])
        elif direction == LEFT:
            errors += 1
            p1.append(GAPCHAR)
            j -= 1
            p2.append(query[j])
        elif direction == UP:
            i -= 1
            p1.append(ref[i])
            p2.append(GAPCHAR)
            errors += 1
        else:
            assert False, "DP table corrupt"

    start1 = i if start_in_ref else 0
    start2 = j if start_in_query else 0

    errors += (i - start1) + (j - start2)

    while j > 0:
        p1.append(GAPCHAR)
        j -= 1
        p2.append(query[j])
    while i > 0:
        i -= 1
        p1.append(ref[i])
        p2.append(GAPCHAR)
    assert i == 0 and j == 0

    p1 = "".join(p1[::-1])
    p2 = "".join(p2[::-1])

    return Alignment(p1, p2, start1, best_i, start2, best_j, best_score, errors)


def align_global(ref: str, query: str, match=1, mismatch=-2, insertion=-2, deletion=-2):
    m = len(ref)
    n = len(query)
    start = 0
    while start < m and start < n and ref[start] == query[start]:
        start += 1

    stop_ref = m
    stop_query = n
    while stop_ref > start and stop_query > start and ref[stop_ref-1] == query[stop_query-1]:
        stop_ref -= 1
        stop_query -= 1

    assert 0 <= start <= stop_ref <= m
    assert 0 <= start <= stop_query <= n

    alignment = align_global_core(
        ref[start:stop_ref], query[start:stop_query], match, mismatch, insertion, deletion)

    alignment.ref_row = ref[:start] + alignment.ref_row + ref[stop_ref:]
    assert alignment.ref_start == 0
    assert alignment.ref_stop == stop_ref - start
    alignment.ref_stop = m
    alignment.query_row = query[:start] + alignment.query_row + query[stop_query:]
    assert alignment.query_start == 0
    assert alignment.query_stop == stop_query - start
    alignment.query_stop = n
    alignment.score += start + m - stop_ref

    return alignment



"""
Compute union of sequences in multiple FASTA files
"""
import logging
from collections import namedtuple
import dnaio

from ..utils import Merger

logger = logging.getLogger(__name__)


def add_arguments(parser):
    arg = parser.add_argument
    # arg('--max-differences', type=int, metavar='MAXDIFF', default=0,
    #     help='Merge sequences if they have at most MAXDIFF differences. '
    #     ' Default: %(default)s')
    arg('fasta', help='FASTA file', nargs='+')


SequenceInfo = namedtuple('SequenceInfo', 'sequence name')


class SequenceMerger(Merger):
    """
    Merge sequences where one is a prefix of the other into single entries.
    """
    def __init__(self):
        super().__init__()

    def merged(self, s, t):
        """
        Merge two sequences if one is the prefix of the other. If they should
        not be merged, None is returned.

        s and t must have attributes sequence and name.
        """
        s_seq = s.sequence
        t_seq = t.sequence
        # Make both sequences the same length - cheap trick to not penalize
        # end gaps
        s_seq += t_seq[len(s_seq):]
        t_seq += s_seq[len(t_seq):]
        if s_seq == t_seq:
            return s
        return None


def main(args):
    merger = SequenceMerger()
    n_read = 0
    for path in args.fasta:
        n = 0
        for record in dnaio.open(path):
            merger.add(SequenceInfo(record.sequence.upper(), record.name))
            n += 1
        n_read += n
        logger.info('Read %s sequences from %s', n, path)
    logger.info('Read %s sequences from %s files', n_read, len(args.fasta))

    for info in merger:
        print('>{}\n{}'.format(info.name, info.sequence))



from igdiscover.align import consensus, multialign


def test_consensus():
    assert consensus((
        "TATTACTGTGCGAG---",
        "TATTACTGTGCGAGAGA",
        "TATTACTGTGCGAGAGA",
        "TATTACTGTGCGAGAG-",
        "TATTACTGTGCGAGAG-",
        "TATTACTGTGCGAG---",
        "TATTACTGTGCGAGA--",
    )) == \
        "TATTACTGTGCGAGAGA"


def test_multialign():
    result = multialign({
        "seq1": "TATTACTGTGCGAG",
        "seq2": "TCTTACGTGCTAG",
    }, program="muscle")
    assert result == {
        "seq1": "TATTACTGTGCGAG",
        "seq2": "TCTTAC-GTGCTAG",
    }



import os
import ruamel.yaml


class ConfigurationError(Exception):
    pass


class Config:
    DEFAULT_PATH = 'igdiscover.yaml'

    def __init__(self, file):
        # Set some defaults.
        self.debug = False
        self.species = None
        self.sequence_type = 'Ig'
        self.merge_program = 'pear'
        self.flash_maximum_overlap = 300
        self.limit = None  # or an integer
        self.multialign_program = 'muscle-fast'
        self.minimum_merged_read_length = 300
        self.mismatch_penalty = None
        self.barcode_length = 0
        self.barcode_consensus = True
        self.iterations = 1
        self.ignore_j = False
        self.d_coverage = 70
        self.subsample = 1000
        self.stranded = False
        self.forward_primers = None
        self.reverse_primers = None
        self.rename = False
        self.race_g = False
        self.seed = 1
        self.exact_copies = None
        self.preprocessing_filter = dict(v_coverage=90, j_coverage=60, v_evalue=1E-3)
        self.pre_germline_filter = dict(
            unique_cdr3s=2,
            unique_js=2,
            whitelist=True,
            cluster_size=0,
            allow_stop=True,
            # allow_chimeras=False,
            cross_mapping_ratio=0.02,
            clonotype_ratio=0.12,
            exact_ratio=0.12,
            cdr3_shared_ratio=0.8,
            unique_d_ratio=0.3,
            unique_d_threshold=10
        )
        self.germline_filter = dict(
            unique_cdr3s=5,
            unique_js=3,
            whitelist=True,
            cluster_size=50,
            allow_stop=False,
            # allow_chimeras=False,
            cross_mapping_ratio=0.02,
            clonotype_ratio=0.12,
            exact_ratio=0.12,
            cdr3_shared_ratio=0.8,
            unique_d_ratio=0.3,
            unique_d_threshold=10
        )
        self.j_discovery = dict(allele_ratio=0.2, cross_mapping_ratio=0.1, propagate=True)
        self.cdr3_location = 'detect'

        self.read_from(file)

    def read_from(self, file):
        content = file.read()
        new_config = self.make_compatible(ruamel.yaml.safe_load(content))
        for key in ('preprocessing_filter', 'pre_germline_filter', 'germline_filter', 'j_discovery'):
            if key in new_config:
                self.__dict__[key].update(new_config[key])
                del new_config[key]
        self.__dict__.update(new_config)

    @staticmethod
    def make_compatible(config):
        """
        Convert old-style configuration to new style.
        Raise ConfigurationError if configuration is invalid.
        Return updated config dict.
        """
        if 'barcode_length' in config:
            raise ConfigurationError(
                'Old-style configuration of barcode length via "barcode_length"'
                'is no longer supported.')
        barcode_length_5prime = config.get('barcode_length_5prime', 0)
        barcode_length_3prime = config.get('barcode_length_3prime', 0)
        if barcode_length_5prime > 0 and barcode_length_3prime > 0:
            raise ConfigurationError(
                'barcode_length_5prime and barcode_length_3prime can currently '
                'not both be greater than zero.')
        if barcode_length_5prime > 0:
            config['barcode_length'] = barcode_length_5prime
        elif barcode_length_3prime > 0:
            config['barcode_length'] = -barcode_length_3prime
        config.pop('barcode_length_5prime', None)
        config.pop('barcode_length_3prime', None)

        if 'seed' in config and config['seed'] is False:
            config['seed'] = None

        for key in ('germline_filter', 'pregermline_filter'):
            if key in config and 'allele_ratio' in config[key]:
                config[key]['clonotype_ratio'] = config[key]['allele_ratio']
        return config

    @classmethod
    def from_default_path(cls):
        with open(cls.DEFAULT_PATH) as f:
            return Config(file=f)


class GlobalConfig:
    def __init__(self):
        self.use_cache = False
        path = os.getenv('XDG_CONFIG_HOME', os.path.expanduser('~/.config'))
        path = os.path.join(path, 'igdiscover.conf')
        if os.path.exists(path):
            with open(path) as f:
                config = ruamel.yaml.safe_load(f)
            if config is None:
                return
            self.use_cache = config.get('use_cache', False)



import random
from igdiscover.trie import Trie
from igdiscover.cli.group import hamming_neighbors
from tinyalign import hamming_distance


def random_nt(length):
    return ''.join(random.choice('ACGT') for _ in range(length))


class TestTrie:
    LENGTHS = (0, 1, 2, 3, 4, 5, 6, 10, 12, 15, 20)

    def setup(self):
        self.strings = set()
        self.trie = Trie()
        # Create a set and a Trie both containing the same strings
        for _ in range(80):
            for length in self.LENGTHS:
                s = random_nt(length)
                self.strings.add(s)
                self.trie.add(s)

    def test_empty_string(self):
        trie = Trie()
        trie.add('')
        assert '' in trie
        assert 'A' not in trie

    def test_contains(self):
        for s in self.strings:
            assert s in self.trie

        for length in self.LENGTHS:
            for _ in range(min(100, 4**length)):
                s = random_nt(length)
                assert (s in self.strings) == (s in self.trie)

    def test_len(self):
        assert len(self.strings) == len(self.trie)

    def naive_has_similar(self, s, distance):
        for t in self.strings:
            if len(t) != len(s):
                continue
            if hamming_distance(t, s) <= distance:
                return True
        return False

    def naive_find_all_similar(self, s, distance):
        for t in self.strings:
            if len(t) != len(s):
                continue
            if hamming_distance(t, s) <= distance:
                yield t

    def test_has_similar(self):
        for s in self.strings:
            assert self.trie.has_similar(s, 0)
            assert self.trie.has_similar(s, 1)
            assert self.trie.has_similar(s, 2)

        for base in self.strings:
            for modified in hamming_neighbors(base):
                assert self.trie.has_similar(modified, 1)

        for errors in range(4):
            for length in self.LENGTHS:
                for _ in range(min(100, 4**length)):
                    s = random_nt(length)
                    assert self.naive_has_similar(s, errors) == self.trie.has_similar(s, errors)

    def test_find_all_similar(self):
        t = Trie()
        t.add('ACGT')
        result = list(t.find_all_similar('ACGG', 1))
        assert set(result) == frozenset(('ACGT',))

        for s in self.strings:
            result = list(self.trie.find_all_similar(s, 0))
            assert result == [s]

            for errors in range(1, 4):
                result = list(self.trie.find_all_similar(s, errors))
                assert s in result

        for base in self.strings:
            for modified in hamming_neighbors(base):
                assert base in self.trie.find_all_similar(modified, 1)

        for errors in range(1, 3):
            for length in self.LENGTHS:
                for _ in range(min(100, 4**length)):
                    s = random_nt(length)
                    expected = set(self.naive_find_all_similar(s, errors))
                    assert expected == set(self.trie.find_all_similar(s, errors))


def main():
    import sys
    n = int(sys.argv[1])
    dist = int(sys.argv[2])

    strings = [random_nt(13) for _ in range(n)]
    print('created random sequences')

    if sys.argv[3] == 'x':
        def naive_has_similar(t):
            for s in strings:
                if hamming_distance(s, t) <= dist:
                    return True
            return False

        hs = 0
        for s in strings:
            hs += int(naive_has_similar(s[1:] + s[0]))
        print('hs:', hs, 'out of', len(strings))

    else:
        trie = Trie()
        for s in strings:
            trie.add(s)
        print('created trie')
        hs = 0
        for s in strings:
            hs += int(trie.has_similar(s[1:] + s[0], dist))

        print('hs:', hs, 'out of', len(strings))


if __name__ == '__main__':
    main()



"""
Dereplicate sequences and remove barcodes

The difference to the 'group' subcommand is that that one
also computes consensus sequences from groups with identical
barcode/CDR3. This one does not.

The barcode can be in the 5' end or the 3' end of the sequence.

Use --trim-g to remove initial runs of G at the 5' end (artifact from RACE protocol).
These are removed after the barcode is removed.
"""

import logging
from collections import defaultdict
from itertools import islice
import json
import dnaio

logger = logging.getLogger(__name__)


def add_arguments(parser):
    arg = parser.add_argument
    arg('--limit', default=None, type=int, metavar='N',
        help='Limit processing to the first N reads')
    arg('--trim-g', action='store_true', default=False,
        help="Trim 'G' nucleotides at 5' end")
    arg('--minimum-length', '-l', type=int, default=0,
        help='Minimum sequence length')
    arg('--barcode-length', '-b', type=int, default=0,
        help="Length of barcode. Positive for 5' barcode, negative for 3' barcode. "
             "Default: %(default)s")
    arg('--json', metavar="FILE", help="Write statistics to FILE")
    arg('fastx', metavar='FASTA/FASTQ',
        help='FASTA or FASTQ file (can be gzip-compressed) with sequences')


def main(args):
    barcode_length = args.barcode_length
    too_short = 0
    n = 0
    sequences = defaultdict(list)  # maps sequences to a list of Sequence objects containing them
    with dnaio.open(args.fastx) as f:
        for record in islice(f, 0, args.limit):
            n += 1
            if len(record) < args.minimum_length:
                too_short += 1
                continue
            sequences[record.sequence].append(record)

    n_written = 0
    for records in sequences.values():
        # If there are multiple records with the same sequence, pick the first
        record = records[0]

        if barcode_length >= 0:
            barcode = record.sequence[:barcode_length]
            unbarcoded = record[barcode_length:]
        else:
            barcode = record.sequence[barcode_length:]
            unbarcoded = record[:barcode_length]

        if args.trim_g:
            # The RACE protocol leads to a run of non-template Gs in the beginning
            # of the sequence, after the barcode.
            unbarcoded.sequence = unbarcoded.sequence.lstrip('G')
            if unbarcoded.qualities:
                unbarcoded.qualities = unbarcoded.qualities[-len(unbarcoded.sequence):]

        name = record.name.split(maxsplit=1)[0]
        if name.endswith(';'):
            name = name[:-1]

        if barcode_length:
            print('>{};barcode={};size={};\n{}'.format(name, barcode, len(records),
                unbarcoded.sequence))
        else:
            print('>{};size={};\n{}'.format(name, len(records),
                unbarcoded.sequence))

        n_written += 1

    logger.info('%s sequences processed', n)
    logger.info('%s sequences long enough', n - too_short)
    logger.info('%s dereplicated sequences written', n_written)

    if args.json:
        stats = {
            'groups_written': n_written,
        }
        with open(args.json, 'w') as f:
            json.dump(stats, f, indent=2)
            print(file=f)



"""
Find common V genes between two different antibody libraries.
"""
import logging
from collections import Counter

from ..table import read_table

logger = logging.getLogger(__name__)


def add_arguments(parser):
    arg = parser.add_argument
    arg('--minimum-frequency', '-n', type=int, metavar='N',
        default=None,
        help='Minimum number of datasets in which sequence must occur (default is no. of files divided by two)')
    arg('table', help='Tables with parsed and filtered IgBLAST results (give at least two)', nargs='+')


def main(args):
    if args.minimum_frequency is None:
        # args.table is a list of file names
        minimum_frequency = max((len(args.table) + 1) // 2, 2)
    else:
        minimum_frequency = args.minimum_frequency
    logger.info('Minimum frequency set to %s', minimum_frequency)

    # Read in tables
    tables = []
    for path in args.table:
        table = read_table(path)
        table = table.loc[:,['V_gene', 'V_SHM', 'V_nt', 'name']]
        tables.append(table)

    # Count V sequence occurrences
    counter = Counter()
    for table in tables:
        counter.update(set(table.V_nt))

    # Find most frequent occurrences and print result
    print('Frequency', 'Gene', '%SHM', 'Sequence', sep='\t')
    for sequence, frequency in counter.most_common():
        if frequency < minimum_frequency:
            break
        names = []
        gene = None
        for table in tables:
            matching_rows = table[table.V_nt == sequence]
            if matching_rows.empty:
                continue
            names.extend(matching_rows.name)
            if gene is None:
                row = matching_rows.iloc[0]
                gene = row['V_gene']
                shm = row['V_SHM']
        print(frequency, gene, shm, sequence, *names, sep='\t')



"""
Compute expression counts

This command takes a table of filtered IgBLAST results (filtered.tab.gz),
filters it by various criteria and then counts how often specific
genes are named.

By default, all genes with non-zero expression are listed, sorted by
name. Use --database to change this.
"""
import sys
import logging
import dnaio
from ..table import read_table
from ..utils import natural_sort_key
from .discoverjd import filter_by_allele_ratio, compute_expressions

logger = logging.getLogger(__name__)


def add_arguments(parser):
    arg = parser.add_argument
    arg('--gene', choices=('V', 'D', 'J'), default='V',
        help='Which gene type: Choose V, D or J. Default: Default: %(default)s')
    arg('--database', metavar='FASTA',
        help='Compute expressions for the sequences that are named in the FASTA '
        'file. Only the sequence names in the file are used! This is the only '
        'way to also include genes with an expression of zero.')
    arg('--plot', metavar='FILE',
        help='Plot expressions to FILE (PDF or PNG)')

    group = parser.add_argument_group('Input table filtering')
    group.add_argument('--d-evalue', type=float, default=None,
        help='Maximal allowed E-value for D gene match. Default: 1E-4 '
        'if --gene=D, no restriction otherwise.')
    group.add_argument('--d-coverage', '--D-coverage', type=float, default=None,
        help='Minimum D coverage (in percent). Default: 70 '
        'if --gene=D, no restriction otherwise.')
    group.add_argument('--d-errors', type=int, default=None,
        help='Maximum allowed D errors. Default: No limit.')

    group = parser.add_argument_group('Expression counts table filtering')
    group.add_argument('--allele-ratio', type=float, metavar='RATIO', default=None,
        help='Required allele ratio. Works only for genes named "NAME*ALLELE". '
        'Default: Do not check allele ratio.')

    arg('table', help='Table with parsed and filtered IgBLAST results')


def plot_counts(counts, gene_type):
    """Plot expression counts. Return a Figure object"""
    from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
    from matplotlib.figure import Figure
    import matplotlib
    import seaborn as sns
    import numpy as np
    sns.set_style('white')
    sns.set_style({'patch.force_edgecolor': False})

    fig = Figure(figsize=((50 + len(counts) * 5) / 25.4, 210/25.4))
    matplotlib.rcParams.update({'font.size': 14})
    FigureCanvas(fig)
    ax = fig.add_subplot(111)
    ax.set_title('{} gene usage'.format(gene_type))
    ax.set_xlabel('{} gene'.format(gene_type))
    ax.set_ylabel('Count')
    ax.set_xticks(np.arange(len(counts)))
    ax.set_xticklabels(counts.index, rotation='vertical')
    ax.grid(axis='y')
    ax.set_xlim(-1, len(counts))
    ax.bar(np.arange(len(counts)), counts['count'])
    fig.set_tight_layout(True)
    return fig


def main(args):
    if args.database:
        with dnaio.open(args.database) as fr:
            gene_names = [record.name for record in fr]
        gene_names.sort(key=natural_sort_key)
    else:
        gene_names = None

    usecols = ['V_gene', 'D_gene', 'J_gene', 'V_errors', 'D_errors', 'J_errors', 'D_covered',
        'D_evalue', 'CDR3_nt']
    table = read_table(args.table, usecols=usecols)
    logger.info('Table with %s rows read', len(table))

    # Set default filters depending on gene
    if args.gene == 'D':
        if args.d_evalue is None:
            args.d_evalue = 1E-4
        if args.d_coverage is None:
            args.d_coverage = 70
    if args.d_evalue is not None:
        table = table[table.D_evalue <= args.d_evalue]
        logger.info('%s rows remain after requiring D E-value <= %s', len(table), args.d_evalue)

    if args.d_coverage is not None:
        table = table[table.D_covered >= args.d_coverage]
        logger.info('%s rows remain after requiring D coverage >= %s', len(table), args.d_coverage)

    if args.d_errors is not None:
        table = table[table.D_errors <= args.d_errors]
        logger.info('%s rows remain after requiring D errors <= %s', len(table), args.d_errors)

    logger.info('Computing expression counts for %s genes', args.gene)
    counts = compute_expressions(table, args.gene)

    # Make sure that always all gene names are listed even if no sequences
    # were assigned.
    if gene_names:
        counts = counts.reindex(gene_names, fill_value=0)

    if args.database and args.allele_ratio:
        logger.error('--database and --allele-ratio cannot be used at the same time.')
        sys.exit(1)

    logger.info('Computed expressions for %d genes', len(counts))
    if args.allele_ratio is not None:
        counts = filter_by_allele_ratio(counts, args.allele_ratio)
        logger.info(
            'After filtering by allele ratio, %d genes remain',
            len(counts))

    # logger.info('%d sequences were not assigned a %s gene', unassigned, args.gene)
    counts.to_csv(sys.stdout, sep='\t')
    logger.info('Wrote expression count table')

    if args.plot:
        plot_counts(counts, args.gene).savefig(args.plot)
        logger.info("Wrote %s", args.plot)



"""
Find V gene sister sequences shared by multiple libraries.
"""
import logging
from collections import Counter
import pandas as pd

logger = logging.getLogger(__name__)


def add_arguments(parser):
    arg = parser.add_argument
    arg('--minimum-frequency', '-n', type=int, metavar='N',
        default=None,
        help='Minimum number of datasets in which sequence must occur (default is no. of files divided by two)')
    arg('--minimum-db-diff', '-b', type=int, metavar='DIST', default=1,
        help='Use only sequences that have at least DIST differences to the database sequence. Default: %(default)s')
    arg('tables', metavar='DISCOVER.TAB',
        help='Table created by the "discover" command (give at least two)', nargs='+')


def main(args):
    if args.minimum_frequency is None:
        minimum_frequency = max((len(args.tables) + 1) // 2, 2)
    else:
        minimum_frequency = args.minimum_frequency
    logger.info('Minimum frequency set to %s', minimum_frequency)

    # Read in tables
    tables = []
    for path in args.tables:
        table = pd.read_csv(path, sep='\t')
        table = table[table.database_diff >= args.minimum_db_diff]
        table = table.dropna()
        tables.append(table)
        if len(table) == 0:
            logger.warn('Table read from %r is empty after filtering out sequences with database diff >= %s.', path, args.minimum_db_diff)

    # Count V sequence occurrences
    counter = Counter()
    for table in tables:
        counter.update(set(table.consensus))

    # Find most frequent occurrences and print result
    print('count', 'gene', 'database_diff', 'sequence', 'names', sep='\t')
    for sequence, frequency in counter.most_common():
        if frequency < minimum_frequency:
            break
        names = []
        gene = None
        for table in tables:
            matching_rows = table[table.consensus == sequence]
            if matching_rows.empty:
                continue
            names.extend(matching_rows.name)
            if gene is None:
                row = matching_rows.iloc[0]
                gene = row.gene
                database_diff = row.database_diff
                #shm = row['V_SHM']
        print(frequency, gene, database_diff, sequence, *names, sep='\t')



"""
Draw a dendrogram of sequences in a FASTA file.
"""
import logging
import numpy as np
import matplotlib
import dnaio
from scipy.spatial import distance
from scipy.cluster import hierarchy
from ..utils import distances


logger = logging.getLogger(__name__)


def add_arguments(parser):
    arg = parser.add_argument
    arg('--mark', '--db', metavar='FASTA',
        help='Path to a FASTA file with a set of "known" sequences. Sequences '
        'in the main file that do *not* occur here will be marked with (new). '
        'If not given, no sequences will be marked (use this to compare two '
        'databases).')
    arg('--method', choices=('single', 'complete', 'weighted', 'average'),
        default='average',
        help='Linkage method. Default: "average" (=UPGMA)')
    arg('fasta', help='Path to input FASTA file')
    arg('plot', help='Path to output PDF or PNG')


class PrefixComparer:
    def __init__(self, sequences):
        self._sequences = [ s.upper() for s in sequences ]

    def __contains__(self, other):
        for seq in self._sequences:
            if seq.startswith(other) or other.startswith(seq):
                return True
        return False


def main(args):
    with dnaio.open(args.fasta) as fr:
        sequences = list(fr)
    logger.info('Plotting dendrogram of %s sequences', len(sequences))
    if args.mark:
        with dnaio.open(args.mark) as fr:
            mark = PrefixComparer(record.sequence for record in fr)
        labels = []
        n_new = 0
        for record in sequences:
            if record.sequence not in mark:
                extra = ' (new)'
                n_new += 1
            else:
                extra = ''
            labels.append(record.name + extra)
        logger.info('%s sequence(s) marked as "new"', n_new)
    else:
        labels = [s.name for s in sequences]

    import seaborn as sns
    import matplotlib.pyplot as plt
    sns.set()
    sns.set_style("white")
    font_size = 297 / 25.4 * 72 / (len(labels) + 5)
    font_size = min(16, max(6, font_size))
    height = font_size * (len(labels) + 5) / 72
    fig = plt.figure(figsize=(210 / 25.4, height))
    matplotlib.rcParams.update({'font.size': 4})
    ax = fig.gca()
    sns.despine(ax=ax, top=True, right=True, left=True, bottom=True)
    sns.set_style('whitegrid')
    if len(sequences) >= 2:
        m = distances([s.sequence for s in sequences])
        y = distance.squareform(m)
        mindist = int(y.min())
        logger.info('Smallest distance is %s. Found between:', mindist)
        for i,j in np.argwhere(m == y.min()):
            if i < j:
                logger.info('%s and %s', labels[i], labels[j])
        l = hierarchy.linkage(y, method=args.method)
        hierarchy.dendrogram(l, labels=labels, leaf_font_size=font_size, orientation='right', color_threshold=0.95*max(l[:,2]))
    else:
        ax.text(0.5, 0.5, 'no sequences', fontsize='xx-large')
    ax.grid(False)
    fig.set_tight_layout(True)
    fig.savefig(args.plot)



#!/usr/bin/env python3
"""
Extract V_SHM values from multiple clonoquery tables
"""
from collections import OrderedDict
import logging
import pandas as pd
import io
import os
import sys
import re


def add_arguments(parser):
	arg = parser.add_argument
	arg('--field', default='V_SHM', help='Which column to extract. Default: %(default)s')
	arg('-n', default=5, type=int, help='Require at least N values in each file for a query to be listed')
	arg('paths', nargs='+')


def read_one_file(path, field_name):
	with open(path) as f:
		header = f.readline()
		sections = re.split(r'\n\t*\n', f.read().strip())
	column_names = header.strip().split('\t')

	queries = dict()
	for section in sections:
		sio = io.StringIO(section)
		n = section.count('# Query: ')
		if n > 100:
			print(repr(section[:10000]))
		query_names = []
		for i in range(n):
			query = sio.readline()
			assert query.startswith('# Query: '), query
			query_name = query.split('\t')[0].partition('# Query: ')[2]
			query_names.append(query_name)
		table = pd.read_csv(sio, header=None, names=column_names, usecols=(field_name,), sep='\t')
		# print('Read a table with {} entries'.format(len(table)), file=sys.stderr)
		for query_name in query_names:
			queries[query_name] = table[field_name]
	return queries


def main(args):
	files = OrderedDict((os.path.basename(path), read_one_file(path, args.field)) for path in args.paths)
	print('read {} files'.format(len(files)), file=sys.stderr)

	query_names = dict()
	for queries in files.values():
		for query_name in queries.keys():
			query_names[query_name] = None
	query_names = list(query_names)

	maxlen = 0
	for queries in files.values():
		maxlen = max(maxlen, max(len(q) for q in queries.values()))

	assert len(set(query_names)) == len(query_names)
	df = pd.DataFrame(index=range(maxlen))
	for query_name in query_names:
		long_enough_in_all_files = all(len(files[file].get(query_name, [])) >= args.n for file in files)
		if not long_enough_in_all_files:
			print('Query {} has too few values in at least one file'.format(query_name), file=sys.stderr)
			continue
		for file in files:
			column_name = '{file}-{query_name}'.format(file=file, query_name=query_name)
			df[column_name] = files[file].get(query_name, pd.Series([]))
	print(df.to_csv(sep='\t', index=False))


if __name__ == '__main__':
	from argparse import ArgumentParser
	parser = ArgumentParser()
	add_arguments(parser)
	args = parser.parse_args()
	main(args)



"""
Filter table with parsed IgBLAST results

Discard the following rows in the table:
- no J assigned
- stop codon found
- V gene coverage less than 90%
- J gene coverage less than 60%
- V gene E-value greater than 1E-3

The filtered table is printed to standard output.
"""
import logging
import json
import pandas as pd

from ..table import fix_columns

logger = logging.getLogger(__name__)


def add_arguments(parser):
    arg = parser.add_argument
    arg('--v-coverage', type=float, default=90, metavar='PERCENT',
        help='Require that the sequence covers at least PERCENT of the V gene. '
        'Default: %(default)s')
    arg('--v-evalue', type=float, default=1E-3, metavar='EVALUE',
        help='Require that the E-value for the V gene match is at most EVALUE. '
        'Default: %(default)s')
    arg('--j-coverage', type=float, default=60, metavar='PERCENT',
        help='Require that the sequence covers at least PERCENT of the J gene. '
        'Default: %(default)s')
    arg('--json', metavar='FILE', help='Write statistics to FILE')
    arg('table', help='Table with filtered IgBLAST results.')


class FilteringStatistics:
    __slots__ = ('total', 'has_vj_assignment', 'has_no_stop', 'good_v_evalue', 'good_v_coverage',
    'good_j_coverage', 'has_cdr3')

    def __init__(self):
        self.total = 0
        self.has_vj_assignment = 0
        self.has_no_stop = 0
        self.good_v_evalue = 0
        self.good_v_coverage = 0
        self.good_j_coverage = 0
        self.has_cdr3 = 0

    def __iadd__(self, other):
        for att in self.__slots__:
            v = getattr(self, att)
            setattr(self, att, v + getattr(other, att))
        return self

    def asdict(self):
        d = dict()
        for att in self.__slots__:
            d[att] = getattr(self, att)
        return d


def filtered_table(table,
        v_gene_coverage,  # at least
        j_gene_coverage,  # at least
        v_gene_evalue,  # at most
    ):
    """
    Discard the following rows in the table:
    - no J assigned
    - stop codon found
    - V gene coverage less than v_gene_coverage
    - J gene coverage less than j_gene_coverage
    - V gene E-value greater than v_gene_evalue

    Return the filtered table.
    """
    stats = FilteringStatistics()
    stats.total = len(table)
    # Both V and J must be assigned
    # (Note V_gene and J_gene columns use empty strings instead of NA)
    filtered = table[(table['V_gene'] != '') & (table['J_gene'] != '')][:]
    stats.has_vj_assignment = len(filtered)
    filtered['V_gene'] = pd.Categorical(filtered['V_gene'])

    # Filter out sequences that have a stop codon
    filtered = filtered[filtered.stop == 'no']
    stats.has_no_stop = len(filtered)

    # Filter out sequences with a too low V gene hit E-value
    filtered = filtered[filtered.V_evalue <= v_gene_evalue]
    stats.good_v_evalue = len(filtered)

    # Filter out sequences with too low V gene coverage
    filtered = filtered[filtered.V_covered >= v_gene_coverage]
    stats.good_v_coverage = len(filtered)

    # Filter out sequences with too low J gene coverage
    filtered = filtered[filtered.J_covered >= j_gene_coverage]
    stats.good_j_coverage = len(filtered)

    stats.has_cdr3 = sum(filtered['CDR3_nt'] != '')
    return filtered, stats


def main(args):
    first = True
    written = 0
    stats = FilteringStatistics()
    for chunk in pd.read_csv(args.table, chunksize=10000, float_precision='high', sep='\t'):
        fix_columns(chunk)
        filtered, chunk_stats = filtered_table(chunk, v_gene_coverage=args.v_coverage,
            j_gene_coverage=args.j_coverage, v_gene_evalue=args.v_evalue)
        stats += chunk_stats
        print(filtered.to_csv(sep='\t', index=False, header=first), end='')
        first = False
        written += len(filtered)

    logger.info('%s rows in input table', stats.total)
    logger.info('%s rows have both V and J assignment', stats.has_vj_assignment)
    logger.info('%s of those do not have a stop codon', stats.has_no_stop)
    logger.info('%s of those have an E-value of at most %s', stats.good_v_evalue, args.v_evalue)
    logger.info('%s of those cover the V gene by at least %s%%', stats.good_v_coverage, args.v_coverage)
    logger.info('%s of those cover the J gene by at least %s%%', stats.good_j_coverage, args.j_coverage)
    logger.info('%d rows written', written)
    logger.info('%s rows have a recognized CDR3 (these are not filtered)', stats.has_cdr3)

    if args.json:
        with open(args.json, 'w') as f:
            json.dump(stats.asdict(), f, indent=2)
            print(file=f)
